#!/usr/bin/env python2.6
################################################################################
#
#       This file is part of Gato (Graph Animation Toolbox) 
#
#   file:   GatoExport.py
#   author: Alexander Schliep (alexander@schliep.org)
#
#       Copyright (C) 2010, Alexander Schliep, Winfried Hochstaettler and 
#       Copyright 1998-2001 ZAIK/ZPR, Universitaet zu Koeln
#                                   
#       Contact: alexander@schliep.org, winfried.hochstaettler@fernuni-hagen.de
#
#       Information: http://gato.sf.net
#
#       This library is free software; you can redistribute it and/or
#       modify it under the terms of the GNU Library General Public
#       License as published by the Free Software Foundation; either
#       version 2 of the License, or (at your option) any later version.
#
#       This library is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#       Library General Public License for more details.
#
#       You should have received a copy of the GNU Library General Public
#       License along with this library; if not, write to the Free
#       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1303.5  USA
#
#
#
#
#       This file is version $Revision: 353 $ 
#                       from $Date: 2010-06-03 14:58:37 -0400 (Thu, 03 Jun 2010) $
#             last change by $Author: schliep $.
#
################################################################################
import os
import StringIO
import tokenize
import re
import pdb
from bs4 import BeautifulSoup
from math import sqrt, pi, sin, cos, atan2, degrees, log10, floor, ceil
from WebGatoJS import animationhead

#Global constants for tokenEater
line_count = 1
keywordsList = [
          "del", "from", "lambda", "return",
          "and", "elif", "global", "not", "try",
          "break", "else", "if", "or", "while",
          "class", "except", "import", "pass",
          "continue", "finally", "in", "print",
          "def", "for", "is", "raise"]

operatorsList = ["+", "-", "*", "/", "^", "%", "=",
                 "+=", "-=", "*=", "/=", "^=", "%=",
                 ">", "<", "==", "!=", ">=", "<=",
                 "**", "<>", "|", "&", "<<", ">>",
                 "//", "~", "**="]

specialList = ["(", ",", ".", ")", "[", "]"]

begun_line = False
num_spaces = 0.0
algo_lines = []     # tokenEater uses this to write algorithm lines to
prev = ["",-1]
indent_stack = [0]
last_line = ''
id_prefixes = ['g1_', 'g2_']
svg_drop_shadow = '''
    <filter id="dropshadow" height="130%">
      <feGaussianBlur in="SourceAlpha" stdDeviation="1"/> <!-- stdDeviation is how much to blur -->
        <feOffset dx="2.5" dy="2.5" result="offsetblur"/> <!-- how much to offset -->
        <feMerge> 
          <feMergeNode/> <!-- this contains the offset blurred image -->
          <feMergeNode in="SourceGraphic"/> <!-- this contains the element that the filter is applied to -->
       </feMerge>
    </filter>
'''


def tokenEater(type, token, (srow, scol), (erow, ecol), line):
    global line_count
    global prev
    global begun_line
    global num_spaces
    global algo_lines
    global indent_stack
    global last_line

    if begun_line and last_line != line:
        # This handles the case where there is a multi-line algorithm command 
        # that spans the lines with a backslash(look at line 10/11 of Prim)
        line_count += 1
        algo_lines.append('</text>\n')
        algo_lines.append('<text blank = "false" id="%s" class="code_line" x="0" y="0" dx="%d" text-anchor="start" '\
                       'fill="black" font-family="Courier New" font-size="14.0" font-style="normal" style="cursor: pointer">' % ("l_" + str(line_count), 8*indent_stack[len(indent_stack)-1]))

    indent_const = 22
    if (type == 0): #EOF.  Reset globals
        line_count = 1
        num_spaces = 0.0
        indent_stack = [0]
        begun_line = False
        prev = ["",-1]
    elif (type == 1): #Word.  Potential keyword.  Must check keywordsList
        if begun_line == False:
            begun_line = True
            algo_lines.append('<text blank = "false" id="%s" class="code_line" x="0" y="0" dx="%d" text-anchor="start" '\
                       'fill="black" font-family="Courier New" font-size="14.0" font-style="normal" style="cursor: pointer">' % ("l_" + str(line_count), 4*indent_stack[len(indent_stack)-1]))

        if token in keywordsList:
            if (prev[1] == -1 or prev[1] == 5 or prev[1] == 4 or prev[1] == 54 or prev[1] == 6): #first word
                algo_lines.append('<tspan font-weight="bold">%s</tspan>' % (token))
            elif (prev[0] in specialList and (prev[0] != "]" and prev[0] != ")")):
                algo_lines.append('<tspan font-weight="bold">%s</tspan>' % token)
            else:
                algo_lines.append('<tspan font-weight="bold"> %s</tspan>' % token)
        else:
            if (prev[1] == -1 or prev[1] == 5 or prev[1] == 4 or prev[1] == 54 or prev[1] == 6): #first word
                algo_lines.append('%s' % (token))
            elif (prev[0] in specialList and (prev[0] != "]" and prev[0] != ")")):
                algo_lines.append('%s' % token)
            else:
                algo_lines.append(' %s' % token)
    elif (type == 4): #Newline on nonempty line
        algo_lines.append('</text>\n')
        begun_line = False
        line_count += 1
    elif (type == 5):  #Arbitrary number of tabs at beginning of line  tabs are 4 spaces long
        num_spaces = 0.0
        for x in token:
            if ord(x) == 9:
                num_spaces += 7.7
            elif ord(x) == 32:
                num_spaces += 1.0

        num_spaces = int(floor(num_spaces))
        indent_stack.append(num_spaces)
        #num_spaces = int(floor(len(token))/4)
    elif (type == 6):  #One backpedal
        indent_stack.pop()
    elif (type == 51): #Operators and punctuation
        if begun_line == False:
            begun_line = True
            algo_lines.append('<text blank = "false" id="%s" class="code_line" x="0" y="0" dx="%d" text-anchor="start" '\
                       'fill="black" font-family="Courier New" font-size="14.0" font-style="normal" style="cursor: pointer">' % ("l_" + str(line_count), 4*indent_stack[len(indent_stack)-1]))

        if token in operatorsList:
            if token == "<":
                token = "&lt;"
            elif token == "<<":
                token = "&lt;&lt;"
            elif token == "<=":
                token = "&lt;="
            elif token == "<>":
                token = "&lt;>"

            if (prev[1] == -1 or prev[1] == 5 or prev[1] == 4 or prev[1] == 54 or prev[1] == 6): #first word
                algo_lines.append('%s' % (token))
            else:
                algo_lines.append(' %s' % token)
        else:
            if (prev[1] == -1 or prev[1] == 5 or prev[1] == 4 or prev[1] == 54 or prev[1] == 6): #first word
                algo_lines.append('%s' % (token))
            elif prev[0] in operatorsList:
                algo_lines.append('%s' % token)
            else:
                algo_lines.append('%s' % token)
    elif (type == 53): #Comment
        if begun_line == False:
            begun_line = True
            algo_lines.append('<text blank = "false" id="%s" class="code_line" x="0" y="0" dx="%d" text-anchor="start" '\
                       'fill="black" font-family="Courier New" font-size="14.0" font-style="normal" style="cursor: pointer">' % ("l_" + str(line_count), 4*indent_stack[len(indent_stack)-1]))

        if (prev[0] in specialList and (prev[0] != "]" and prev[0] != ")")):
            algo_lines.append('%s' % token)
        else:
            algo_lines.append('%s' % token)
    elif (type == 54): #Empty line with newline
        algo_lines.append('<text blank = "true" id="%s" class="code_line" x="0" y="0" dx="%d" text-anchor="start" '\
                       'fill="black" font-family="Courier New" font-size="14.0" font-style="normal" style="cursor: pointer"></text>\n' % ("l_" + str(line_count), 4*indent_stack[len(indent_stack)-1]))
        line_begun = False
        line_count += 1
    else:
        if begun_line == False:
            begun_line = True
            algo_lines.append('<text blank = "false" id="%s" class="code_line" x="0" y="0" dx="%d" text-anchor="start" '\
                       'fill="black" font-family="Courier New" font-size="14.0" font-style="normal" style="cursor: pointer">' % ("l_" + str(line_count), 4*indent_stack[len(indent_stack)-1]))

        if (prev[0] in specialList and (prev[0] != "]" and prev[0] != ")")):
            algo_lines.append('%s' % token)
        else:
            algo_lines.append(' %s' % token)
    
    last_line = line
    if type != 0:
        prev[0] = token
        prev[1] = type


def cmd_as_javascript(cmd, idPrefix=''):
    """ Return a list of methodname, target and args """
    def quote(s, prefix=''):
        return "\"%s\"" % (prefix+str(s))
    
    if len(cmd.target) == 1:
        target = quote(cmd.target[0], idPrefix)
    else:
        target = quote(cmd.target, idPrefix)
        
    result = [cmd.time, cmd.method.__name__, target]
    for arg in cmd.args:
        result.append(quote(arg))
    
    # Special case for some animatin commands
    if cmd.method.__name__ == 'SetAllVerticesColor' and 'vertices' in cmd.kwargs:
        for v in cmd.kwargs['vertices']:
            result.append(quote(v))

    return result

def change_id_format(field):
    def edge_replace(match):
        return match.group().replace(' ', '').replace('(', '').replace(')', '').replace(',', '-')
    edge_re = re.compile(r'g[12]_\(\d+, \d+\)')
    if edge_re.search(field):
        return edge_re.sub(edge_replace, field)
    else:
        return field

def collectAnimations(histories, prefixes):
    """ Given a list of animation histories (aka list of AnimationCommands)
        combine them, giving all targets of animation commands their history-
        specific prefix, sort them and return a list of JavaScripts arrays.
    """ 
    mergedCmds = [cmd_as_javascript(cmd, prefixes[0]) for cmd in histories[0]]
    for i, h in enumerate(histories[1:]):
        mergedCmds += [cmd_as_javascript(cmd, prefixes[i+1]) for cmd in h]
    mergedCmds = sorted(mergedCmds, key=lambda cmd: cmd[0])
    # Replace absolute times by duration
    currentTime = mergedCmds[0][0]
    for i, cmd in enumerate(mergedCmds):
        duration = max(1,int(round((cmd[0] - currentTime) * 1000, 0)))
        currentTime = cmd[0]
        mergedCmds[i][0] = str(duration)
        # We want to change the edge ids to a different form
        for j in xrange(2, len(mergedCmds[i])):
            if 'Edge' in mergedCmds[i][1]:
                mergedCmds[i][j] = change_id_format(mergedCmds[i][j])
    return ["Array(" + ", ".join(cmd) + ")" for cmd in mergedCmds]

def get_edge_id(v, w, idPrefix):
    return idPrefix + '{}-{}'.format(v, w)


def get_graph_as_svg_str_standalone(graphDisplay, x_add, y_add, file, idPrefix='', translate=None):
    ''' Returns a 3-tuple of (svg_string, width, height).  The SVG string returned
        is designed for standalone viewing.
    '''
    def update_min_and_max(x, y, min_x, min_y, max_x, max_y):
        if min_x == None:
            min_x = x
        else:
            min_x = min(x, min_x)
        if max_x == None:
            max_x = x
        else:
            max_x = max(x, max_x)
        if min_y == None:
            min_y = y
        else:
            min_y = min(y, min_y)
        if max_y == None:
            max_y = y
        else:
            max_y = max(y, max_y)
        return min_x, min_y, max_x, max_y

    ret_strs = []
    min_x, max_x, min_y, max_y = None, None, None, None

    # Keep track of vertex_radius as we have to shift everything over by that much so 
    # vertices on the far left don't appear half off the screen
    vertex_radius = None
    for v in graphDisplay.G.Vertices():
        x,y,r = graphDisplay.VertexPositionAndRadius(v)
        if vertex_radius == None:
            vertex_radius = r
            x_add -= vertex_radius
            y_add -= vertex_radius
            break
    if translate:
        ret_strs.append('<g transform="translate(%d %d)">\n' % (translate[0], translate[1]))

    # Write Edges:
    for v,w in graphDisplay.G.Edges():
        vx, vy, r = graphDisplay.VertexPositionAndRadius(v)
        wx, wy, r = graphDisplay.VertexPositionAndRadius(w)
        col = graphDisplay.GetEdgeColor(v,w)
        width = graphDisplay.GetEdgeWidth(v,w)
        
        vy -= y_add
        wy -= y_add
        vx -= x_add
        wx -= x_add
        min_x, min_y, max_x, max_y = update_min_and_max(vx, vy, min_x, min_y, max_x, max_y)
        min_x, min_y, max_x, max_y = update_min_and_max(wx, wy, min_x, min_y, max_x, max_y)

        if graphDisplay.G.directed == 0:
            ret_strs.append('<line x1="%s" y1="%s" x2="%s" y2="%s" stroke="%s"'\
                       ' stroke-width="%s"/>\n' % (vx, vy, wx, wy, col, width))
        else:
            x1,y1,x2,y2 = graphDisplay.directedDrawEdgePoints(graphDisplay.VertexPosition(v),
                graphDisplay.VertexPosition(w), 0)
            
            x1e,y1e = graphDisplay.CanvasToEmbedding(x1,y1)
            x2e,y2e = graphDisplay.CanvasToEmbedding(x2,y2)
            y1e -= y_add
            y2e -= y_add
            x1e -= x_add
            x2e -= x_add
            min_x, min_y, max_x, max_y = update_min_and_max(x1e, y1e, min_x, min_y, max_x, max_y)
            min_x, min_y, max_x, max_y = update_min_and_max(x2e, y2e, min_x, min_y, max_x, max_y)

            if graphDisplay.G.QEdge(w,v): # Directed edges both ways
                ret_strs.append('<line x1="%s" y1="%s" x2="%s" y2="%s" stroke="%s"'\
                           ' stroke-width="%s"/>\n' % (x1e, y1e, x2e, y2e, col, width))
            else: # Just one directed edge
                # XXX How to color arrowhead?
                l = sqrt((float(wx)-float(vx))**2 + (float(wy)-float(vy))**2)
                if (l < .001):
                    l = .001

                c = (l-2*graphDisplay.zVertexRadius)/l + .01
                tmpX = float(vx) + c*(float(wx) - float(vx))
                tmpY = float(vy) + c*(float(wy) - float(vy))
                
                #dx = 0 #offset of wx to make room for arrow
                #dy = 0 #offset of wy
                cr = 0
                #Took out marker-end="url(#Arrowhead)" and added polyline
                #Shrink line to make room for arrow
                for z in graphDisplay.G.Vertices():
                    cx,cy,cr = graphDisplay.VertexPositionAndRadius(z)
                    cy -= y_add
                    cx -= x_add
                    min_x, min_y, max_x, max_y = update_min_and_max(cx, cy, min_x, min_y, max_x, max_y)
                    if(cx == wx and cy == wy):
                        angle = atan2(int(float(wy))-int(float(vy)), int(float(wx))-int(float(vx)))
                        ret_strs.append('<line x1="%s" y1="%s" x2="%f" y2="%f" stroke="%s"'\
                               ' stroke-width="%s" />\n' % (vx, vy, tmpX, tmpY,
                                                            col, width))
                        break

                #Temporary settings for size of polyline arrowhead
                a_width = (1 + 1.5/(1*pow(log10(float(width)), 6)))
                if(a_width > 5.0):
                    a_width = 5.0
                a_width *= float(width) 
                p1 = (0,0)
                p2 = (0, a_width)
                p3 = (cr, a_width/2)
                angle = degrees(atan2(int(wy)-int(vy), int(wx)-int(vx)))
                c = (l-2*graphDisplay.zVertexRadius)/l
                tmpX = float(vx) + c*(float(wx) - float(vx))
                tmpY = float(vy) + c*(float(wy) - float(vy))
                ret_strs.append('<polyline points="%f %f %f %f %s %f" fill="%s" transform="translate(%f,%f)'\
                           ' rotate(%f %f %f)" />\n' % (p1[0], p1[1], p2[0], p2[1], p3[0], p3[1],
                                                        col, tmpX, tmpY - a_width/2, angle, p1[0], a_width/2))


        # Write Edge Annotations
        if graphDisplay.edgeAnnotation.QDefined((v,w)):
            da = graphDisplay.edgeAnnotation[(v,w)]
            x,y = graphDisplay.canvas.coords(graphDisplay.edgeAnnotation[(v,w)])
            xe,ye = graphDisplay.CanvasToEmbedding(x,y)
            y -= y_add
            ye -= y_add
            x -= x_add
            xe -= x_add
            min_x, min_y, max_x, max_y = update_min_and_max(x, y, min_x, min_y, max_x, max_y)
            min_x, min_y, max_x, max_y = update_min_and_max(xe, ye, min_x, min_y, max_x, max_y)
            text = graphDisplay.canvas.itemcget(graphDisplay.edgeAnnotation[(v,w)],"text") 
            size = r * 0.9
            offset = 0.33 * size
            col = 'black'
            if text != "":
                ret_strs.append('<text x="%s" y="%s" text-anchor="center" '\
                           'fill="%s" font-family="Helvetica" '\
                           'font-size="%s" font-style="normal">%s</text>\n' % (ye+offset,col,size,text))

    for v in graphDisplay.G.Vertices():
        x,y,r = graphDisplay.VertexPositionAndRadius(v)
        y -= y_add
        x -= x_add
        min_x, min_y, max_x, max_y = update_min_and_max(x, y, min_x, min_y, max_x, max_y)

        # Write Vertex
        col = graphDisplay.GetVertexColor(v)
        fw = graphDisplay.GetVertexFrameWidth(v)
        fwe,dummy = graphDisplay.CanvasToEmbedding(fw,0)
        stroke = graphDisplay.GetVertexFrameColor(v)

        ret_strs.append('<circle cx="%s" cy="%s" r="%s" fill="%s" stroke="%s"'\
                   ' stroke-width="%s" style="filter:url(#dropshadow)"/>\n' % (x,y,r,col,stroke,fwe))

        # Write Vertex Label
        col = graphDisplay.canvas.itemcget(graphDisplay.drawLabel[v], "fill")
        size = r*1.0
        offset = 0.33 * size

        ret_strs.append('<text x="%s" y="%s" text-anchor="middle" fill="%s" font-family="Helvetica" '\
                   'font-size="%s" font-style="normal" font-weight="bold" >%s</text>\n' % (x, y+offset,col,size, graphDisplay.G.GetLabeling(v)))
        # Write vertex annotation
        #size = r*0.9
        size = 14
        text = graphDisplay.GetVertexAnnotation(v)
        col = 'black'
        if text != "":
            ret_strs.append('<text x="%s" y="%s" text-anchor="left" fill="%s" font-weight="bold" font-family="Helvetica" '\
                       'font-size="%s" font-style="normal">%s</text>\n' % (x+r+1,y+r*1.5+2.5,col,size,text))

    if translate:
        ret_strs.append('</g>')

    width = max_x-min_x+vertex_radius*2
    height = max_y-min_y+vertex_radius*2
    return '\n'.join(ret_strs), width, height



def get_graph_as_svg_str_for_animation(graphDisplay, x_add, y_add, file, idPrefix=''):
    ''' Returns an svg string.  The SVG string returned 
        is designed for use with javascript with WebGato
    '''
    ret_strs = []

    # Write Edges
    for v,w in graphDisplay.G.Edges():
        vx,vy,r = graphDisplay.VertexPositionAndRadius(v)
        wx,wy,r = graphDisplay.VertexPositionAndRadius(w)
        col = graphDisplay.GetEdgeColor(v,w)
        width = graphDisplay.GetEdgeWidth(v,w)
        
        vy -= y_add
        wy -= y_add
        vx -= x_add
        wx -= x_add
        
        edge_id = get_edge_id(v, w, idPrefix)
        ret_strs.append('<g id="%s" class="edge_group" style="cursor: pointer">' % (edge_id + '_group'))
        if graphDisplay.G.directed == 0:
            ret_strs.append('<line id="%s" class="edge" x1="%s" y1="%s" x2="%s" y2="%s" stroke="%s"'\
                       ' stroke-width="%s"/>\n' % (edge_id, vx, vy, wx, wy, col, width))
        else:
            x1,y1,x2,y2 = graphDisplay.directedDrawEdgePoints(graphDisplay.VertexPosition(v),
                graphDisplay.VertexPosition(w), 0)
            
            x1e,y1e = graphDisplay.CanvasToEmbedding(x1,y1)
            x2e,y2e = graphDisplay.CanvasToEmbedding(x2,y2)
            y1e -= y_add
            y2e -= y_add
            x1e -= x_add
            x2e -= x_add

            if graphDisplay.G.QEdge(w,v): # Directed edges both ways
                ret_strs.append('<line id="%s" class="edge" x1="%s" y1="%s" x2="%s" y2="%s" stroke="%s"'\
                           ' stroke-width="%s"/>\n' % (edge_id, x1e, y1e, x2e, y2e, col, width))
            else: # Just one directed edge
                # XXX How to color arrowhead?
                l = sqrt((float(wx)-float(vx))**2 + (float(wy)-float(vy))**2)
                if (l < .001):
                    l = .001

                c = (l-2*graphDisplay.zVertexRadius)/l + .01
                tmpX = float(vx) + c*(float(wx) - float(vx))
                tmpY = float(vy) + c*(float(wy) - float(vy))
                
                #dx = 0 #offset of wx to make room for arrow
                #dy = 0 #offset of wy
                cr = 0
                #Took out marker-end="url(#Arrowhead)" and added polyline
                #Shrink line to make room for arrow
                for z in graphDisplay.G.Vertices():
                    cx,cy,cr = graphDisplay.VertexPositionAndRadius(z)
                    cy -= y_add
                    cx -= x_add
                    if(cx == wx and cy == wy):
                        angle = atan2(int(float(wy))-int(float(vy)), int(float(wx))-int(float(vx)))
                        ret_strs.append('<line id="%s" class="edge" x1="%s" y1="%s" x2="%f" y2="%f" stroke="%s"'\
                               ' stroke-width="%s" />\n' % (edge_id, vx, vy, tmpX, tmpY,
                                                            col, width))
                        break

                #Temporary settings for size of polyline arrowhead
                a_width = (1 + 1.5/(1*pow(log10(float(width)), 6)))
                if(a_width > 5.0):
                    a_width = 5.0
                a_width *= float(width) 
                p1 = (0,0)
                p2 = (0, a_width)
                p3 = (cr, a_width/2)
                angle = degrees(atan2(int(wy)-int(vy), int(wx)-int(vx)))
                c = (l-2*graphDisplay.zVertexRadius)/l
                tmpX = float(vx) + c*(float(wx) - float(vx))
                tmpY = float(vy) + c*(float(wy) - float(vy))
                ret_strs.append('<polyline id="%s" class="arrowhead" points="%f %f %f %f %s %f" fill="%s" transform="translate(%f,%f)'\
                           ' rotate(%f %f %f)" />\n' % ('ea' + edge_id, p1[0], p1[1], p2[0], p2[1], p3[0], p3[1],
                                                        col, tmpX, tmpY - a_width/2, angle, p1[0], a_width/2))
        ret_strs.append('</g>')


        # Write Edge Annotations
        if graphDisplay.edgeAnnotation.QDefined((v,w)):
            da = graphDisplay.edgeAnnotation[(v,w)]
            x,y = graphDisplay.canvas.coords(graphDisplay.edgeAnnotation[(v,w)])
            xe,ye = graphDisplay.CanvasToEmbedding(x,y)
            y -= y_add
            ye -= y_add
            x -= x_add
            xe -= x_add

            text = graphDisplay.canvas.itemcget(graphDisplay.edgeAnnotation[(v,w)],"text") 
            size = r * 0.9
            offset = 0.33 * size
            col = 'black'
            if text != "":
                ret_strs.append('<text id="ea%s" class="edge_annotation" x="%s" y="%s" text-anchor="center" '\
                           'fill="%s" font-family="Helvetica" '\
                           'font-size="%s" font-style="normal">%s</text>\n' % (idPrefix+str(xe),
                                                                               ye+offset,col,size,text))

    ret_strs.append('<g id="%s"></g>' % (idPrefix+'pre_vertices_group'))
    for v in graphDisplay.G.Vertices():
        x,y,r = graphDisplay.VertexPositionAndRadius(v)
        y -= y_add
        x -= x_add

        # Write Vertex
        col = graphDisplay.GetVertexColor(v)
        fw = graphDisplay.GetVertexFrameWidth(v)
        fwe,dummy = graphDisplay.CanvasToEmbedding(fw,0)
        stroke = graphDisplay.GetVertexFrameColor(v)

        ret_strs.append('<g id="%s" class="vertex_g" style="cursor: pointer">' % (idPrefix + str(v) + '_group'))
        ret_strs.append('<circle id="%s" class="vertex" cx="%s" cy="%s" r="%s" fill="%s" stroke="%s"'\
                   ' stroke-width="%s" style="filter:url(#dropshadow)"/>\n' % (idPrefix+str(v),x,y,r,col,stroke,fwe))

        # Write Vertex Label
        col = graphDisplay.canvas.itemcget(graphDisplay.drawLabel[v], "fill")
        size = r*1.0
        offset = 0.33 * size

        ret_strs.append('<text id="vl%s" x="%s" y="%s" text-anchor="middle" fill="%s" font-family="Helvetica" '\
                   'font-size="%s" font-style="normal" font-weight="bold" >%s</text>\n' % (idPrefix+str(v),x,
                                                                                           y+offset,col,size,
                                                                                           graphDisplay.G.GetLabeling(v)))
        ret_strs.append('</g>')
        # Write vertex annotation
        #size = r*0.9
        size = 14
        text = graphDisplay.GetVertexAnnotation(v)
        col = 'black'
        if text != "":
            ret_strs.append('<text id="va%s" class="vertex_annotation" x="%s" y="%s" text-anchor="left" fill="%s" font-weight="bold" font-family="Helvetica" '\
                       'font-size="%s" font-style="normal" text_content="%s">%s</text>\n' % (idPrefix+str(v),x+r+1,y+r*1.5+2.5,col,size,text,text))

    return '\n'.join(ret_strs)
    

def compute_coord_changes(gdisp):
    has_elements = False
    x_add, y_add = 0, 0
    for v, w in gdisp.G.Edges():
        vx, vy, r = gdisp.VertexPositionAndRadius(v)
        if not has_elements:
            x_add = vx
            y_add = vy
            has_elements = True
        else:
            x_add = min(vx, x_add)
            y_add = min(vy, y_add)
    for v in gdisp.G.Vertices():
        vx, vy, r = gdisp.VertexPositionAndRadius(v)
        if not has_elements:
            x_add = vx
            y_add = vy
            has_elements = True
        else:
            x_add = min(vx, x_add)
            y_add = min(vy, y_add)
    return x_add, y_add, has_elements

def format_init_edge_infos(info_dict, idPrefix):
    ''' Formats the info_dict to a javascript object '''
    if not info_dict:
        return 'null';
    str_bits = ['{'] # List of strings to return joined at the end(faster than concatenation)
    for tup, info in info_dict.iteritems():
        v, w = tup
        edge_id = get_edge_id(v, w, idPrefix)
        assignment = '"{}": "{}",'.format(edge_id, info)
        str_bits.append(assignment)
    str_bits.append('}')
    return '\n'.join(str_bits)

def format_init_vertex_infos(info_dict, idPrefix):
    ''' Formats the info_dict to a javascript object '''
    if not info_dict:
        return 'null';
    str_bits = ['{'] # List of strings to return joined at the end(faster than concatenation)
    for v, info in info_dict.iteritems():
        vertex_id = idPrefix + str(v)
        assignment = '"{}": "{}",'.format(vertex_id, info)
        str_bits.append(assignment)
    str_bits.append('}')
    return '\n'.join(str_bits)
        
def ExportAlgoInfo(fileName, algorithm):
    if not os.path.exists('./svgs/infos'):
        os.makedirs('./svgs/infos')

    file = open("./svgs/infos/%s" % os.path.basename(fileName).replace("svg", "html"), "w")
    info = algorithm.About()
    soup = BeautifulSoup(info)

    # Link the CSS file
    head = soup.find('head')
    if not head:
        html = soup.find('html')
        html.insert(0, soup.new_tag('head'))
        head = soup.find('head')
    head.insert(0, soup.new_tag('link', rel='stylesheet', href='../css/info_page_style.css'))

    # Insert a div with background-color found by colordef into the <dt> elements
    for dt in soup.find_all('dt'):
        colordef = dt.find('colordef')
        if not colordef:
            continue
        color = colordef['color']
        colordef.extract()
        color_div = soup.new_tag('div')
        color_div.string = '&nbsp;&nbsp;&nbsp;&nbsp;'
        color_div['style'] = 'background-color: %s' % color
        color_div['class'] = 'color_div'
        dt.append(color_div)


    # Add a div with "clear: both" to make the next row
    for dd in soup.find_all('dd'):
        clear_div = soup.new_tag('div')
        clear_div['class'] = 'clear_div'
        dd.insert_after(clear_div)
    
    body = soup.find('body')
    if not body.contents:
        # If there isn't any content add "No algorithm info"
        body.append(soup.new_string('No algorithm info'))

    file.write(soup.prettify(formatter=None))

def format_animation(animation):
    def chunker(seq, size):
        return (seq[pos:pos + size] for pos in xrange(0, len(seq), size))
    if len(animation) < 32000:
        return 'var anim_array = Array(' + ',\n'.join(animation) + ');'
    else:
        ret_str = 'var anim_array = Array(' + ',\n'.join(animation[:32000]) + ');\n'
        for chunk in chunker(animation[32000:], 32000):
            ret_str += 'anim_array.concat(Array(' + ','.join(chunk) + '));\n'
        return ret_str

def construct_title(fileName):
    sp = fileName.split('/')[1].split('--')
    algorithm = sp[0]
    graph = sp[1].split('.')[0]
    return 'Gato -- %s algorithm on %s graph' % (algorithm, graph)

def construct_animation_name(fileName):
    sp = fileName.split('/')[1].split('--')
    algorithm = sp[0]
    graph = sp[1].split('.')[0]
    return '%s algorithm on %s graph' % (algorithm, graph)

def ExportSVG(fileName, algowin, algorithm, graphDisplay, secondaryGraphDisplay=None, 
    secondaryGraphDisplayAnimationHistory=None, showAnimation=False, 
    init_edge_infos=None, init_vertex_infos=None, init_graph_infos=None,
    write_to_png=False, chapter_number=None, algo_div=None):
    """ Export either the current graphs or the complete animation
        (showAnimation=True) to the file fileName.

        showAnimation: We don't do anything if this is False
        init_edge_infos: list of dictionaries with keys that are tuples of form (v1, v2) where v1 and v2 are the vertices
            that an edge connects.  The values keys point to are the initial edge infos of that edge.
    """
    global algo_lines

    algo_lines = []
    file = open(fileName, 'w')

    # Figure out how much we want to pull the graph to the left and top before we reset the graph
    end_g1_x_add, end_g1_y_add, end_g1_has_elements = compute_coord_changes(graphDisplay)
    end_g2_x_add, end_g2_y_add = 0, 0
    if secondaryGraphDisplay:
        end_g2_x_add, end_g2_y_add, end_g2_has_elements = compute_coord_changes(secondaryGraphDisplay)

    if showAnimation:
        try:
            if secondaryGraphDisplayAnimationHistory:
                animation = collectAnimations([algorithm.animation_history.getHistoryOne(),
                                               secondaryGraphDisplayAnimationHistory.getHistoryTwo(),
                                               algowin.codeLineHistory],
                                              ['g1_', 'g2_', 'l_'])
            else:
                animation = collectAnimations([algorithm.animation_history.getHistoryOne(),
                                               algowin.codeLineHistory],
                                              ['g1_', 'l_'])
        except IndexError as e:
            print "Error:"
            print e
            print "Filename: ", fileName
            print "Algowin: ", algowin
            print "Algorithm: ", algorithm
            print "graphDisplay: ", graphDisplay
            return

    # Reload the graph and execute prolog so we can save the initial state to SVG
    algorithm.Start(prologOnly=True)

    start_g1_x_add, start_g1_y_add, start_g1_has_elements = compute_coord_changes(graphDisplay)
    if start_g1_has_elements and end_g1_has_elements:
        g1_x_add, g1_y_add = min(start_g1_x_add, end_g1_x_add), min(start_g1_y_add, end_g1_y_add)
    elif start_g1_has_elements:
        g1_x_add, g1_y_add = start_g1_x_add, start_g1_y_add
    else:
        g1_x_add, g1_y_add = end_g1_x_add, end_g1_y_add

    start_g2_x_add, start_g2_y_add, g2_x_add, g2_y_add = 0, 0, 0, 0
    if secondaryGraphDisplay:
        start_g2_x_add, start_g2_y_add, start_g2_has_elements = compute_coord_changes(secondaryGraphDisplay)
        if start_g2_has_elements and end_g2_has_elements:
            g2_y_add, g2_x_add = min(start_g2_y_add, end_g2_y_add), min(start_g2_x_add, end_g2_x_add)
        elif start_g2_has_elements:
            g2_y_add, g2_x_add = start_g2_y_add, start_g2_x_add
        else:
            g2_y_add, g2_x_add = end_g2_y_add, end_g2_x_add

    if showAnimation and not write_to_png:
        # Build the SVG graph string
        graph_strs = []
        graph_type = "undirected" if graphDisplay.G.directed == 0 else "directed"
        graph_strs.append('<g id="g1" type="%s" style="visibility: hidden">\n' % (graph_type))

        g1_str = get_graph_as_svg_str_for_animation(graphDisplay, g1_x_add, g1_y_add, file, idPrefix=id_prefixes[0])
        graph_strs.append(g1_str)
        graph_strs.append('</g>\n')
        if secondaryGraphDisplay:
            graph_type = "undirected" if secondaryGraphDisplay.G.directed == 0 else "directed"
            graph_strs.append('<g id="g2" type="%s" style="visibility: hidden">\n' % (graph_type))
            g2_str = get_graph_as_svg_str_for_animation(secondaryGraphDisplay, g2_x_add, g2_y_add, file, idPrefix=id_prefixes[1])
            graph_strs.append(g2_str)
            graph_strs.append('</g>\n')

        # Build the Algorithm SVG string
        source = algorithm.GetSource()
        tokenize.tokenize(StringIO.StringIO(source.replace('\\', '\\\\')).readline, tokenEater)
        algowin.CommitStop()

        # Merge the animation into the HTML
        str_vars = {
            'title': construct_title(fileName),
            'animation_name': construct_animation_name(fileName),
            'chapter_number': chapter_number or 0,
            'algo_div': algo_div or '',
            'info_file': 'infos/' + fileName[fileName.rindex('/') + 1:], 
            'animation': format_animation(animation),
            'graph_str': '\n'.join(graph_strs), 
            'algo_str': '<g id="codelines" style="visibility: hidden">' + ''.join(algo_lines) + "</g>",
            'g1_x_add': g1_x_add,
            'g1_y_add': g1_y_add,
            'g2_x_add': g2_x_add,
            'g2_y_add': g2_y_add,
            'g1_init_edge_info': format_init_edge_infos(init_edge_infos[0], id_prefixes[0]) if init_edge_infos else 'null',
            'g2_init_edge_info': format_init_edge_infos(init_edge_infos[1], id_prefixes[1]) if init_edge_infos and len(init_edge_infos) > 1 else 'null',
            'g1_init_graph_info': '"%s"' % init_graph_infos[0] if init_graph_infos and init_graph_infos[0] else '""',
            'g2_init_graph_info': '"%s"' % init_graph_infos[1] if init_graph_infos and init_graph_infos[1] else '""',
            'g1_init_vertex_info': format_init_vertex_infos(init_vertex_infos[0], id_prefixes[0]) if init_vertex_infos else 'null',
            'g2_init_vertex_info': format_init_vertex_infos(init_vertex_infos[1], id_prefixes[1]) if init_vertex_infos else 'null',
        }
        file.write(animationhead % str_vars)
        file.close()

        # Export the algorithm info to its own HTML file        
        ExportAlgoInfo(fileName, algorithm)


    elif not showAnimation and not write_to_png:
        edge_padding = 5
        g1_svg_body_str, g1_width, g1_height = get_graph_as_svg_str_standalone(graphDisplay, g1_x_add, 
            g1_y_add, file, idPrefix=id_prefixes[0], translate=(edge_padding, edge_padding))
        g2_svg_body_str, g2_width, g2_height = '', 0, 0
        g2_y_padding = 0
        if secondaryGraphDisplay:
            g2_y_padding = 10
            g2_svg_body_str, g2_width, g2_height = get_graph_as_svg_str_standalone(secondaryGraphDisplay, g2_x_add, 
                g2_y_add, file, idPrefix=id_prefixes[1], translate=(edge_padding,g1_height+g2_y_padding+edge_padding))

        width = int(max(g1_width, g2_width))+edge_padding*2
        height = int(g1_height+g2_height+g2_y_padding)+edge_padding*2 
        scale = min(200.0/width, 1.0)
        
        # Put a border between the graphs
        if secondaryGraphDisplay:
            y = g1_height + g2_y_padding/2 + edge_padding
            g2_svg_body_str += '<line x1="0" x2="%d" y1="%d" y2="%d" stroke="#000" stroke-width="3.0" />' % (width, y, y)
        svg_str = '\n'.join(['<svg width="%d" height="%d">' % (width, height), svg_drop_shadow, g1_svg_body_str, g2_svg_body_str, '</svg>'])
        file.write(svg_str)
        file.close()

    elif not showAnimation and write_to_png:
        '''
        The commands on Ubuntu to get these libraries installed:
        sudo apt-get install libcairo2-dev librsvg2-dev python-rsvg
        '''
        import cairo
        import rsvg

        file.close()

        edge_padding = 15
        g1_svg_body_str, g1_width, g1_height = get_graph_as_svg_str_standalone(graphDisplay, g1_x_add, 
            g1_y_add, file, idPrefix=id_prefixes[0], translate=(edge_padding, edge_padding))
        g2_svg_body_str, g2_width, g2_height = '', 0, 0
        g2_y_padding = 0
        if secondaryGraphDisplay:
            g2_y_padding = 10
            g2_svg_body_str, g2_width, g2_height = get_graph_as_svg_str_standalone(secondaryGraphDisplay, g2_x_add, 
                g2_y_add, file, idPrefix=id_prefixes[1], translate=(edge_padding,g1_height+g2_y_padding+edge_padding))

        width = int(max(g1_width, g2_width))+edge_padding*2
        height = int(g1_height+g2_height+g2_y_padding)+edge_padding*2 
        scale = min(min(200.0/width, 1.0), min(200.0/height, 1.0))
        
        # Put a border between the graphs
        if secondaryGraphDisplay:
            y = g1_height + g2_y_padding/2 + edge_padding
            g2_svg_body_str += '<line x1="0" x2="%d" y1="%d" y2="%d" stroke="#000" stroke-width="3.0" />' % (width, y, y)
        svg_str = '\n'.join(['<svg width="%d" height="%d">' % (width, height), svg_drop_shadow, g1_svg_body_str, g2_svg_body_str, '</svg>'])

        img = cairo.ImageSurface(cairo.FORMAT_ARGB32, int(ceil(width*scale)), int(ceil(height*scale)))

        ctx = cairo.Context(img)
        ctx.scale(scale, scale)
        handler= rsvg.Handle(None, svg_str)
        handler.render_cairo(ctx)
        img.write_to_png(fileName)
        return {'width': width*scale, 'height': height*scale}




################################################################################
#
#       This file is part of Gato (Graph Animation Toolbox) 
#
#	file:   Graph.py
#	author: Alexander Schliep (alexander@schliep.org)
#
#       Copyright (C) 1998-2015, Alexander Schliep, Winfried Hochstaettler and 
#       Copyright 1998-2001 ZAIK/ZPR, Universitaet zu Koeln
#                                   
#       Contact: alexander@schliep.org, winfried.hochstaettler@fernuni-hagen.de             
#
#       Information: http://gato.sf.net
#
#       This library is free software; you can redistribute it and/or
#       modify it under the terms of the GNU Library General Public
#       License as published by the Free Software Foundation; either
#       version 2 of the License, or (at your option) any later version.
#
#       This library is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#       Library General Public License for more details.
#
#       You should have received a copy of the GNU Library General Public
#       License along with this library; if not, write to the Free
#       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
#       This file is version $Revision: 670 $ 
#                       from $Date: 2015-01-13 16:04:11 -0500 (Tue, 13 Jan 2015) $
#             last change by $Author: schliep $.
#
################################################################################

from GatoGlobals import *
from DataStructures import Point2D, VertexLabeling, EdgeLabeling, EdgeWeight
import GraphUtil #import ConnectedComponents, FindBipartitePartitions 


from math import sqrt
import logging
import string
log = logging.getLogger("Graph.py")

################################################################################
#
# Graph
#
################################################################################
class Graph:
    """ Provides a mathematical graph object consisting of vertices 
        and (directed) edges connecting those vertices. Graphs have
    
        - a labeling for vertices allowing to specify names
    
        - an embedding of vertices into 2D-space
    
        - one or more sets of edge weights  
    
        Vertices are specified via id (integer number) and edges via
        (tail,head)-tuples 
    
        NOTE: ids are supposed to be consecutive and ranging from 0
        to G.Order() - 1 !!! Use the labeling to *display* other numbers
        for vertices.
    
        At least one set of edge weights is assumed to exist and accessible
        as self.edgeWeights[0]; self.euclidian and Euclidify refer to this
        self.edgeWeights[0]
    
    """
    
    def __init__(self):
        self.simple           = 1
        self.euclidian        = 1
        self.directed         = 0
        self.vertices         = [] 
        self.adjLists         = {}
        self.invAdjLists      = {}   # Inverse Adjazenzlisten
        self.highVertexID     = 0    # INTERNAL
        self.embedding        = VertexLabeling() # 2D-Positions
        self.labeling         = VertexLabeling() # Names of vertices
        self.edgeWeights      = {}   # Dictionary of edge labellings
        self.edgeWeights[0]   = EdgeWeight(self)
        self.vertexWeights    = {}   # None by default
        self.size             = 0    # Number of edges
        self.edgeWidth        = None
        self.vertexAnnotation = None
        self.edgeAnnotation   = None
        self.properties       = {}
        
    def AddVertex(self):
        """ Add an isolated vertex. Returns the id of the new vertex """
        id = self.GetNextVertexID()	
        self.vertices.append(id)
        self.adjLists[id]    = []
        self.invAdjLists[id] = []
        return id
        
    def DeleteVertex(self, v):
        """ Delete the vertex v and its incident edges """
        outVertices = self.OutNeighbors(v)[:] # Need a copy here
        inVertices = self.InNeighbors(v)[:]
        for w in outVertices:
            self.DeleteEdge(v,w)
        for w in inVertices:
            if w != v: # We have already deleted loops
                self.DeleteEdge(w,v)
        self.vertices.remove(v)
        #self.adjLists[v] = None
        #self.invAdjLists[v] = None 
        # XXX Should clean up all other stuff too ...
        
        
    def QVertex(self, v):
        """ Check whether v is a vertex """
        return v in self.vertices
        
    def AddEdge(self, tail, head, initialize_weight=True):
        """ Add an edge (tail,head). Returns tail, head. See SubGraph.AddEdge
            for an explanation.
            Raises GraphNotSimpleError if
            - trying to add a loop
            - trying to add an edge multiply 
        
            In case of directed graphs (tail,head) and (head,tail)
            are distinct edges """
        if self.simple == 1 and tail == head: # Loop
            raise GraphNotSimpleError('(%d,%d) is a loop' % (tail,head))
        if self.directed == 0 and tail in self.adjLists[head]: 
            raise GraphNotSimpleError('(%d,%d) is already an undirected edge' % (head,tail))
        if head in self.adjLists[tail]: # Multiple edge
            raise GraphNotSimpleError('(%d,%d) is already an directed edge' % (tail,head))
            
        self.adjLists[tail].append(head)
        self.invAdjLists[head].append(tail)
        self.size = self.size + 1

        if not initialize_weight:
            return tail, head

        if self.QEuclidian():
            t = self.GetEmbedding(tail)
            h = self.GetEmbedding(head)
            self.SetEdgeWeight(0,tail,head,sqrt((h.x - t.x)**2 + (h.y - t.y)**2))
        else:
            self.SetEdgeWeight(0,tail,head,0)
        for i in xrange(1,self.NrOfEdgeWeights()):
            self.SetEdgeWeight(i,tail,head,0)
        return tail, head
        
        
    def DeleteEdge(self,tail,head):
        """ Deletes edge (tail,head). Does *not* handle undirected graphs
            implicitely. Raises NoSuchEdgeError upon error. """
        
        try:
            self.adjLists[tail].remove(head)
            self.invAdjLists[head].remove(tail)
            self.size = self.size - 1
        except KeyError:
            raise NoSuchEdgeError("(%d,%d) is not an edge." % (tail,head))
            
            
    def Edge(self,tail,head):
        """  Handles undirected graphs by return correct ordered
             vertices as (tail,head). Raises NoSuchEdgeError upon error. """
        
        if tail not in self.vertices or head not in self.vertices:
            raise NoSuchEdgeError("(%d,%d) is not an edge." % (tail,head))
            
        if head in self.adjLists[tail]:
            return (tail,head)
        elif self.directed == 0 and tail in self.adjLists[head]:
            return (head,tail)
        else:
            raise NoSuchEdgeError("(%d,%d) is not an edge." % (tail,head))
            
            
    def QEdge(self,tail,head):
        """ Returns 1 if (tail,head) is an edge in G. If G is undirected
            order of vertices does not matter """
        if self.directed == 1:	
            return head in self.adjLists[tail]
        else: 
            return (head in self.adjLists[tail]) or (tail in self.adjLists[head])


    def QEdgeWidth(self):
        """ Returns 1 if individual edge widths are defined, 0 else """
        return self.edgeWidth != None
        

    def EdgeWidth(self, tail, head):
        return self.edgeWidth[(tail,head)]
    
            
    def Neighborhood(self,v):
        """ Returns the vertices which are connected to v. Does handle
            undirected graphs (i.e., returns vertices w s.t. either 
            (v,w) or (w,v) is an edge) """
        
        if self.directed:
            return self.OutNeighbors(v)
        else:
            return self.InOutNeighbors(v)
            
            
    def InNeighbors(self,v):
        """ Returns vertices w for which (w,v) is an edge """
        return self.invAdjLists[v]
        
        
    def OutNeighbors(self,v):
        """ Returns vertices w for which (v,w) is an edge """
        return self.adjLists[v]
        
        
    def InOutNeighbors(self,v):
        """ Returns vertices w for which (v,w) or (w,v) is an edge """	
        return self.InNeighbors(v) + self.OutNeighbors(v)
        
        
    def InEdges(self,v):
        """ Returns edges (*,v) """	
        f = lambda x, vertex = v : (x,vertex)
        return map(f, self.invAdjLists[v])
        
        
    def OutEdges(self,v):
        """ Returns edges (v,*) """	
        f = lambda x, vertex = v : (vertex,x)
        return map(f ,self.adjLists[v])
        
        
    def IncidentEdges(self,v):
        """ Returns edges (v,*) and (*,v) """	
        return self.InEdges(v) + self.OutEdges(v)
        
        
    def Edges(self):
        """ Returns all edges """		
        tmp = []
        for v in self.vertices:
            tmp = tmp + self.OutEdges(v)
        return tmp
        
    def Vertices(self):
        """ Returns all edges """		
        return self.vertices
        
    def printMy(self):
        """ Debugging only """
        for v in self.vertices:
            print v, " -- ", self.adjLists[v]
            
            
    def GetNextVertexID(self):
        """ *Internal* returns next free vertex id """
        self.highVertexID = self.highVertexID + 1
        return self.highVertexID
        
        
    def Order(self):
        """ Returns order i.e., the number of vertices """
        return len(self.vertices)
        
        
    def Size(self):
        """ Returns size i.e., the number of edge """
        return self.size 
        
        
    def Degree(self, v):
        """ Returns the degree of the vertex v, which is
            - the number of incident edges in the undirect case
            - the number of outgoing edges in the directed case """
        
        if self.directed:
            return len(self.adjLists[v])
        else:
            return len(self.adjLists[v]) + len(self.invAdjLists[v])
            
    def InDegree(self, v):
        """ Returns the number of incoming edges for direct graphs """
        if self.directed:
            return len(self.invAdjLists[v])
        else:
            return None # Proper error to raise?
            
            
    def OutDegree(self, v):
        """ Returns the number of incoming edges for direct graphs """
        if self.directed:
            return len(self.adjLists[v])
        else:
            return None # Proper error to raise?
            
            
    def QEuclidian(self):
        """ Returns 1 if the graph is euclidian, 0 else """
        return self.euclidian
        
        
    def QDirected(self):
        """ Returns 1 if the graph is directed, 0 else """
        return self.directed


    def QIsolated(self, v):
        """ Returns 1 if the vertex v is isolated"""
        return (len(self.adjLists[v]) == 0) and (len(self.invAdjLists[v]) == 0)
        
        
    def CalculateWidthFromWeight(self, scale, weightID = 0):
        """ Calculate width of edges (self.edgeWidth will be used by 
            GraphDisplay if not none) from the specified set of edge
            weights. 
        
            Default: weightID = 0 is used """
        
        self.edgeWidth = EdgeLabeling()
        edges = self.Edges()
        if len(edges) > 0:
            maxWeight = max(self.edgeWeights[weightID].label.values())
            for e in edges:
                self.edgeWidth[e] = max(
                    scale * (1 + 35 * self.edgeWeights[weightID][e] / maxWeight),
                    0.001)
        
    def NrOfEdgeWeights(self):
        return len(self.edgeWeights.keys())

    def SetEdgeWeight(self,i,v,w,value):
        self.edgeWeights[i][(v,w)] = value

    def GetEdgeWeight(self,i,v,w):
        return self.edgeWeights[i][(v,w)]

    def NrOfVertexWeights(self):
        return len(self.vertexWeights.keys())

    def SetVertexWeight(self,i,v,value):
        self.vertexWeights[i][v] = value

    def GetVertexWeight(self,i,v):
        return self.vertexWeights[i][v]

    def GetLabeling(self,v):
        return self.labeling[v]
    
    def SetLabeling(self,v, value):
        self.labeling[v] = value

    def GetEmbedding(self,v):
        return self.embedding[v]
    
    def SetEmbedding(self,v, x, y):
        self.embedding[v] = Point2D(x,y)
                  
    def Euclidify(self):
        """ Replace edge weights with weightID = 0 with Euclidean distance 
            between incident vertices """
        
        for v in self.vertices:
            for w in self.adjLists[v]:
                d = ((self.embedding[v].x - self.embedding[w].x)**2 + 
                     (self.embedding[v].y - self.embedding[w].y)**2)**(.5)
                
                if self.edgeWeights[0].QInteger():
                    self.edgeWeights[0][(v,w)] = int(round(d))
                else:
                    self.edgeWeights[0][(v,w)] = d
                    
        self.euclidian = 1
        
        
    def Integerize(self, weightID = 0):
        """ Integerize: Make all edge weights integers """
        
        if weightID == 'all':
            for w in self.edgeWeights.keys():
                self.edgeWeights[w].Integerize()
        else:
            self.edgeWeights[weightID].Integerize()
            
            
    def Undirect(self):
        """ If (u,v) and (v,u) are edges in the directed graph, remove one of them.
            to make graph undirected (no multiple edges allowed). Which one gets
            deleted depends on ordering in adjacency lists. """
        if not self.directed:
            return
            
        for v in self.vertices:
            for w in self.adjLists[v]:
                if v in self.adjLists[w]:
                    self.DeleteEdge(w,v)
                    
        self.directed = 0
        
    def SetProperty(self, name, val):
        """ Set the value of property 'name' to 'val' """
        self.properties[name] = val
        
    def Property(self,name):
        """ Return the value of property 'name'. If the property
           'name' has not been set 'Unknown' is returned
           
            XXX the catbox format does not support storing properties
        """
        if name == "Directed":
            return self.directed
        elif name == "Undirected":
            return not self.directed
        elif name == "EdgeWeights":
            return self.NrOfEdgeWeights()
        elif name == "Euclidean":
            return self.euclidian       
        elif name == "Simple":
            return self.simple
        elif name == "VertexWeights":
            return self.NrOfVertexWeights()
        elif name == "Connected":
            # NOTE: Connected only for Undirected graphs
            if self.directed:
                return 0
            else:
                return self.QConnected()
        elif name == "Bipartite":
            # NOTE: Bipartite only for Undirected graphs
            if self.directed:
                return 0
            else:
                return self.QBipartite()
        elif name == "EvenOrder":
            return (self.Order() % 2) == 0
        elif name == 'NonNegativeEdgeWeights':
            # There is a bug in gred which allows editing edge weights of
            # Euclidean graphs.
            #if self.euclidian:
            #    return True
            #else:
            return self.QNonNegativeEdgeWeights() 
        elif name == 'Size':
            return self.size
        else:
            try:
                return self.properties[name]
            except KeyError:
                return 'Unknown'

    def QConnected(self):
        """ Check if I am connected (assuming I am undirected)"""
        components = GraphUtil.ConnectedComponents(self)
        return len(components) == 1

    def QBipartite(self):
        """ Check if I am bipartite (assuming I am undirected)"""
        partitions = GraphUtil.FindBipartitePartitions(self)
        return len(partitions[0]) > 0 and len(partitions[1]) > 0 


    def QNonNegativeEdgeWeights(self):
        """ Check that all edge-weights of edgeWeights[0] are non-negative """
        if self.NrOfEdgeWeights() > 0:
            for e in self.Edges():
                if self.edgeWeights[0][e] < 0:
                    return False
        return True
            
    def About(self, graphName=""):
        """ Return string containing HTML code providing information
            about the graph

        """
        unknownProps = []
        knownProps = ""
        for name in  gProperty.keys():
            value = self.Property(name)
            if value == 'Unknown':
                unknownProps.append(gProperty[name][2])
            else:
                if gProperty[name][0] == 0:
                    if value == 1:
                        knownProps += "<li> The graph is %s.\n" % gProperty[name][2]
                    else:
                        knownProps += "<li> The graph is not %s.\n" % gProperty[name][2]
                else:
                        knownProps += "<li> The graph has %s %s.\n" % (str(value), gProperty[name][2])
                    
        result = """<HTML><BODY>
        <H3>Graph information</H3>

        <p>The graph <tt>%s</tt> has order %d and size %d and has the following
        properties:
        <ul>
        %s
        </ul>
        </p>

        <p>The properties %s are unknown or undetermined.</p>
        </BODY></HTML>
        """ % (graphName, self.Order(), self.Size(), knownProps, string.join(unknownProps,', '))
        return result




        
        ################################################################################
        #
        # Induced Subgraph
        #
        ################################################################################
        
class SubGraph(Graph):
    """ Provides a subgraph, i.e., a subset of the vertices and edges 
        of a specified graph
    
        Vertices are specified via ids from its supergraph and edges via
        (tail,head)-tuples 
    
        It also keeps track of the subgraphs total weight (= sum of edge 
        weights) for weights with weightID == 0
    """
    
    
    def __init__(self,G):
        Graph.__init__(self)
        self.superGraph    = G
        
        self.embedding     = self.superGraph.embedding
        self.labeling      = self.superGraph.labeling
        self.edgeWeights   = self.superGraph.edgeWeights
        
        self.directed = self.superGraph.directed
        
        self.totalWeight   = 0
        
        
    def AddVertex(self,v):
        """ Add a vertex from the supergraph to the subgraph.
            Returns NoSuchVertexError if v does not exist in
            supergraph """
        try:
            if not v in self.vertices:
                self.vertices.append(v)
                #f = lambda x, vertexList=self.vertices: x in vertexList
                #self.adjLists[v]    = filter(f, self.superGraph.adjLists[v])
                #self.invAdjLists[v] = filter(f, self.superGraph.invAdjLists[v])
                self.adjLists[v] = []
                self.invAdjLists[v] = []
        except:
            raise NoSuchVertexError("%d is not a vertex in the supergraph" % v)
            
    def AddEdge(self,tail,head):
        """ Add an edge from the supergraph to the subgraph. Note: For undirected
            graphs the edge (tail, head) might actually be (head, tail), so we
            return the correct tail, head to caller.
            
            Will also add tail and/or head if there are not already in subgraph """
        try:
            if not self.directed:
                tail, head = self.superGraph.Edge(tail,head) 
            if not tail in self.vertices:
                self.AddVertex(tail)
            if not head in self.vertices:
                self.AddVertex(head)
            self.adjLists[tail].append(head)
            self.invAdjLists[head].append(tail)
            self.size = self.size + 1
            try:
                w = self.superGraph.edgeWeights[0][(tail,head)]
            except KeyError:
                w = 0.0 # XXX we dont have w weight for the edge. Make totalWeight configurable/subclass
            self.totalWeight += w
            return tail, head
            
        except (KeyError, NoSuchVertexError, NoSuchEdgeError):
            raise NoSuchEdgeError("(%d,%d) is not an edge in the supergraph." % (tail,head))
            
    def AddSubGraph(self,G):
        """ Add subgraph G to self. Will do nothing if self and G 
            have distinct supergraphs """
        if self.superGraph != G.superGraph:
            log.error("AddSubGraph: distinct superGraphs")
            return
        for e in G.Edges():
            self.AddEdge(e[0],e[1])
        for v in G.vertices:
            self.AddVertex(v)
            
            
    def DeleteEdge(self,tail,head):
        """ Delete edge from subgraph. Raises NoSuchEdgeError
            upon error """
        if tail in self.vertices and head in self.vertices:
            superEdge = self.superGraph.Edge(tail,head)
            self.totalWeight =  self.totalWeight - self.superGraph.edgeWeights[0][superEdge]
            self.adjLists[tail].remove(head)
            self.invAdjLists[head].remove(tail)
            self.size = self.size - 1
        else:
            raise NoSuchEdgeError("(%d,%d) is not an edge." % (tail,head))
            
    def Clear(self):
        """ Delete all vertices and edges from the subgraph. """
        self.vertices         = [] 
        self.adjLists         = {}
        self.invAdjLists      = {}   # Inverse Adjazenzlisten
        self.size = 0
        self.totalWeight   = 0
        
        
    def GetNextVertexID(self):
        """ *Internal* safeguard """
        log.error("Induced Subgraph -> GetNextVertexID should never have been called")
        
    def Weight(self):
        """ Returns the total weight (= sum of edge weights) of subgraph """
        return self.totalWeight
        
        
    def QEuclidian(self):
        """ Returns 1 if the super graph is euclidian, 0 else """
        return self.superGraph.euclidian
        
        
    def QDirected(self):
        """ Returns 1 if the super graph is directed, 0 else"""
        return self.superGraph.directed
        
    def QEdge(self,tail,head):
        """ Returns 1 if (tail,head) is an edge in G """
        if not tail in self.vertices or not head in self.vertices:
            return 0
        if self.directed == 1:	
            return head in self.adjLists[tail]
        else: 
            return head in self.adjLists[tail] or tail in self.adjLists[head]
            
            



################################################################################
#
#       This file is part of Gato (Graph Animation Toolbox) 
#
#	file:   GatoConfiguration.py
#	author: Alexander Schliep (alexander@schliep.org)
#
#       Copyright (C) 1998-2015, Alexander Schliep, Winfried Hochstaettler and 
#       Copyright 1998-2001 ZAIK/ZPR, Universitaet zu Koeln
#                                   
#       Contact: alexander@schliep.org, winfried.hochstaettler@fernuni-hagen.de             
#       Information: http://gato.sf.net
#
#       This library is free software; you can redistribute it and/or
#       modify it under the terms of the GNU Library General Public
#       License as published by the Free Software Foundation; either
#       version 2 of the License, or (at your option) any later version.
#
#       This library is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#       Library General Public License for more details.
#
#       You should have received a copy of the GNU Library General Public
#       License along with this library; if not, write to the Free
#       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
#       This file is version $Revision: 670 $ 
#                       from $Date: 2015-01-13 16:04:11 -0500 (Tue, 13 Jan 2015) $
#             last change by $Author: schliep $.
#
###############################################################################

import ConfigParser
import tkSimpleDialog 
import sys
import os
import tkFont
from Tkinter import *
from EditObjectAttributesDialog import TkIntEntry, TkStringPopupSelector, TkColorSelector

import StringIO

default_cfg = """
[animation]
BlinkRate = 10
BlinkRepeat = 3

[algorithm]
AlgoFont = Courier
AlgoFontSize = 10
# FG determines the foreground color and BG the background
# Valid colors are Tk names or #rrggbb hex tripel
AlgoFG = black 
AlgoBG = white 
BreakpointFG = #ff0000
BreakpointBG = #e5e5e5
InteractiveFG = #009900
InteractiveBG = #e5e5e5
ActiveFG = black
ActiveBG = #bbbbff
"""

class GatoConfiguration:
    """ GatoConfiguration provides a collection of all editable
    configuration items
    
    Configurations are read from
    - default_cfg string
    - gato.cfg
    - ~user/.gato.cfg
    """
    
    def __init__(self, parent):
        self.parent = parent
        self.config = ConfigParser.ConfigParser()
        self.config.readfp(StringIO.StringIO(default_cfg))
        self.config.read(['gato.cfg', os.path.expanduser('~/.gato.cfg')])
        # Keys which have widgets for editing
        self.editkeys = ['blinkrate', 'blinkrepeat', 
                         'algofont', 'algofontsize', 
                         'algofg', 'algobg',
                         'breakpointfg', 'breakpointbg',
                         'interactivefg', 'interactivebg',
                         'activefg', 'activebg']
        # setup a list of legal config keys and remember what section
        # they are from 
        self.keys = []
        self.section = {}
        for sec in self.config.sections():
            self.keys += self.config.options(sec)
            for opt in self.config.options(sec):
                self.section[opt] = sec
                #self.fonts = ['Courier', 'Helvetica']
        self.fonts = tkFont.families(self.parent)
        self.modified = 0
        
    def writeBack(self):
        """ write the current configuration to the user's config file """
        file = os.path.expanduser('~/.gato.cfg')
        try:
            cf = open(file, 'w')
        except:
            return
            
        self.config.write(cf)
        cf.close()
        
        
    def get(self, name):
        if name in self.keys:
            return self.config.get(self.section[name], name)
        else:
            raise 'NoOptionError'
            
    def set(self, name, value):
        if name in self.keys:
            self.config.set(self.section[name], name, "%s" % value)
            self.modified = 1
        else:
            raise 'NoOptionError'
            
            
    def edit(self):
        """ Bring up the editor for the configuration """
        conf_dialog = EditPreferencesDialog(self.parent, self)
        if self.modified == 1:
            self.writeBack()
            self.parent.SetFromConfig()
            
class EditPreferencesDialog(tkSimpleDialog.Dialog):
    def __init__(self, master, gatoconfig):
        self.widgets = {}
        self.gatoconfig = gatoconfig
        tkSimpleDialog.Dialog.__init__(self, master, "Preferences")
        
        
    def pane(self, master, name, desc):
        if desc == "":
            frame = LabelFrame(master, relief=RIDGE, borderwidth=2, text=name)
        else:
            frame = LabelFrame(master, relief=RIDGE, borderwidth=2, text=name + ":" + desc)            
        frame.pack(expand=1, fill=BOTH, side=TOP)
        result = Frame(frame)
        return result
        
    def body(self, master):
        self.resizable(0,0)	
        
        #------------------ pane ------------------------------------------
        pane = self.pane(master, "Animation", "")
        row = 0
        Label(pane, text="Delay between instructions:").grid(row=row, 
                                                             column=0, sticky=E)	
        self.widgets['blinkrate'] = TkIntEntry(pane, 3)
        self.widgets['blinkrate'].tkWidget().grid(row=row, column=1, sticky=W)
        self.widgets['blinkrate'].set(100)
        
        row += 1
        Label(pane, text="Number of blinks:").grid(row=row, column=0, sticky=E)	
        self.widgets['blinkrepeat'] = TkIntEntry(pane, 2)
        self.widgets['blinkrepeat'].tkWidget().grid(row=row, column=1, sticky=W)
        self.widgets['blinkrepeat'].set(4)
        
        pane.pack(expand=1, fill=BOTH, padx=4, pady=4, side=BOTTOM)
        
        #------------------ pane ------------------------------------------
        pane = self.pane(master, "Algorithm Font", "")
        row = 1
        Label(pane, text="Font:").grid(row=row, column=0, sticky=E)	
        self.widgets['algofont'] = TkStringPopupSelector(pane, self.gatoconfig.fonts)
        self.widgets['algofont'].tkWidget().grid(row=row, column=1, 
                                                 sticky=W, columnspan=2)
        
        row += 1
        Label(pane, text="Font Size:").grid(row=row, column=0, sticky=E)	
        self.widgets['algofontsize'] = TkIntEntry(pane, 2)
        self.widgets['algofontsize'].tkWidget().grid(row=row, column=1, sticky=W)
        self.widgets['algofontsize'].set(12)
        
        pane.pack(expand=1, fill=BOTH, padx=4, pady=4, side=BOTTOM)
        #------------------ pane ------------------------------------------
        pane = self.pane(master, "Algorithm Colors", "")
        
        row = 1
        Label(pane, text="Foreground").grid(row=row, column=1, sticky=W)
        Label(pane, text="Background").grid(row=row, column=2, sticky=W)
        
        row += 1
        Label(pane, text="Font Color:").grid(row=row, column=0, sticky=E)	
        self.widgets['algofg'] = TkColorSelector(pane, 'black')
        self.widgets['algofg'].tkWidget().grid(row=row, column=1, sticky=W)
        self.widgets['algobg'] = TkColorSelector(pane, 'black')
        self.widgets['algobg'].tkWidget().grid(row=row, column=2, sticky=W)
        
        row += 1
        Label(pane, text="Breakpoint Font Color:").grid(row=row, column=0, sticky=E)	
        self.widgets['breakpointfg'] = TkColorSelector(pane, 'black')
        self.widgets['breakpointfg'].tkWidget().grid(row=row, column=1, sticky=W)
        self.widgets['breakpointbg'] = TkColorSelector(pane, 'black')
        self.widgets['breakpointbg'].tkWidget().grid(row=row, column=2, sticky=W)
        
        row += 1
        Label(pane, text="Interactive Font Color:").grid(row=row, column=0, sticky=E)	
        self.widgets['interactivefg'] = TkColorSelector(pane, 'black')
        self.widgets['interactivefg'].tkWidget().grid(row=row, column=1, sticky=W)
        self.widgets['interactivebg'] = TkColorSelector(pane, 'black')
        self.widgets['interactivebg'].tkWidget().grid(row=row, column=2, sticky=W)
        
        row += 1
        Label(pane, text="Active lines:").grid(row=row, column=0, sticky=E)	
        self.widgets['activefg'] = TkColorSelector(pane, 'black')
        self.widgets['activefg'].tkWidget().grid(row=row, column=1, sticky=W)
        self.widgets['activebg'] = TkColorSelector(pane, 'black')
        self.widgets['activebg'].tkWidget().grid(row=row, column=2, sticky=W)
        
        pane.pack(expand=1, fill=BOTH, padx=4, pady=4, side=BOTTOM)
        
        for pref in self.gatoconfig.editkeys:
            self.widgets[pref].set(self.gatoconfig.get(pref))
            
    def validate(self):
        ## Colors need no validating
    
        # Everything is valid
        for pref in self.gatoconfig.editkeys:
            self.gatoconfig.set(pref, self.widgets[pref].get())
        return 1
        
        
        
        
        #---------------- Test code ---------------------------------------------
class TkTestFrame(Frame):

    def __init__(self, parent=None):
        Frame.__init__(self,parent)
        Pack.config(self)
        self.createWidgets()
        #self.fonts = tkFont.families(self)
        self.config = GatoConfiguration(self) 
        self.fonts = ['Helvetica','Times']
        
    def createWidgets(self):
        self.QUIT = Button(self, text='QUIT', foreground='red', 
                           command=self.quit)
        self.QUIT.pack(side=LEFT)
        self.About = Button(self, text='Preferences', foreground='red', 
                           command=self.About)
        self.About.pack(side=LEFT)
        
    def About(self):
        self.config.edit()
        
        
if __name__ == '__main__':
    app = TkTestFrame()
    app.mainloop()
    
    
    
    



#!/usr/bin/env python
################################################################################
#
#       This file is part of Gato (Graph Algorithm Toolbox) 
#       You can find more information at 
#       http://gato.sf.net
#
#	file:   EditObjectAttributesDialog.py
#	author: Alexander Schliep (alexander@schliep.org)
#
#       Copyright (C) 1998-2015, Alexander Schliep, Winfried Hochstaettler and 
#       Copyright 1998-2001 ZAIK/ZPR, Universitaet zu Koeln
#                                   
#       Contact: alexander@schliep.org, winfried.hochstaettler@fernuni-hagen.de             
#
#       Information: http://gato.sf.net
#
#       This library is free software; you can redistribute it and/or
#       modify it under the terms of the GNU Library General Public
#       License as published by the Free Software Foundation; either
#       version 2 of the License, or (at your option) any later version.
#
#       This library is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#       Library General Public License for more details.
#
#       You should have received a copy of the GNU Library General Public
#       License along with this library; if not, write to the Free
#       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
#       This file is version $Revision: 670 $ 
#                       from $Date: 2015-01-13 16:04:11 -0500 (Tue, 13 Jan 2015) $
#             last change by $Author: schliep $.
#
################################################################################
from Tkinter import *
from ScrolledText import *
import tkSimpleDialog 
import tkMessageBox
from tkColorChooser import askcolor
import copy
import sys
import os
import types


def typed_assign(var, val):
    result = type(var)(val)
    result.__dict__ = copy.copy(var.__dict__)
    return result
    
    
    
    #-------------------------------------------------------------------------------
class TkBoolEntry:
    """Tk entry field for selecting one Bool"""

    def __init__(self, master, default=False):
        
        self.frame = Frame(master, relief=FLAT)
        self.status = IntVar()
        self.status.set(default)
        useDefaultButton = Checkbutton(self.frame, variable=self.status, command=self.toggle)
        useDefaultButton.grid(row=0, column=0, padx=4, pady=3, sticky=W)
        
    def tkWidget(self):
        return self.frame
        
    def get(self):
        return bool(self.status)
        
    def set(self, value):
        self.status = value

    def toggle(self):
        self.status = not self.status


class TkStringEntry:
    """Tk entry field for editing strings"""
    
    def __init__(self, master, width):
        self.entryWidget = Entry(master, width=width, exportselection=FALSE)
        
    def tkWidget(self):
        return self.entryWidget
        
    def get(self):
        return self.entryWidget.get()
        
    def set(self, value):
        self.entryWidget.delete(0,END)
        self.entryWidget.insert(0,"%s" % value)
        
    def select(self):    
        self.entryWidget.selection_range(0,"end")
        self.entryWidget.focus_set()
        
        
class TkIntEntry(TkStringEntry):
    """Tk entry field for editing one integer"""
    
    def get(self):
        return int(self.entryWidget.get())
        
        
class TkFloatEntry(TkStringEntry):
    """Tk entry field for editing one float"""
    
    def get(self):
        return float(self.entryWidget.get())
        
        
class TkDefaultMixin:
    """Mixin for TkStringEntry, TkIntEntry, TkFloatEntry, ... to deal with
       values which have an externally defined default value. Combination
       of 'use default' checkbox and corresponding entry field """
    
    def __init__(self, master, useDefault, defaultValue, defaultText=None):
        if defaultText is None:
            defaultText="Use default ("+str(defaultValue)+")"
        self.frame = Frame(master, relief=FLAT)
        self.useDefault = IntVar()
        self.useDefault.set(useDefault)
        self.defaultValue = defaultValue
        useDefaultButton = Checkbutton(self.frame, text=defaultText,
                                       variable=self.useDefault,
                                       command=self.toggleDefault)
        useDefaultButton.grid(row=0, column=0, padx=4, pady=3, sticky=W)
        
    def finish(self):
        self.entryWidget.grid(row=0, column=1, padx=4, pady=3, sticky=W)
        self.switchDefault(self.useDefault.get())   
        
    def UseDefault(self):
        return self.useDefault.get()
        
    def switchDefault(self, value):
        if value == 0:
            self.entryWidget['state'] = NORMAL
            self.entryWidget.delete(0,END)
            self.set(self.defaultValue)
        else:
            self.entryWidget.delete(0,END)
            self.entryWidget['state'] = DISABLED
            
    def toggleDefault(self):
        self.switchDefault(self.useDefault.get())            
        
        
class TkDefaultStringEntry(TkStringEntry, TkDefaultMixin):

    def __init__(self, master, width, useDefault, defaultValue):
        TkDefaultMixin.__init__(self, master, useDefault, defaultValue)
        TkStringEntry.__init__(self, self.frame, width)
        self.finish()        
        
    def tkWidget(self): # To avoid ambiguity
        return self.frame
        
        
class TkDefaultIntEntry(TkIntEntry, TkDefaultMixin):

    def __init__(self, master, width, useDefault, defaultValue):
        TkDefaultMixin.__init__(self, master, useDefault, defaultValue)
        TkIntEntry.__init__(self, self.frame, width)
        self.finish()        
        
    def tkWidget(self): # To avoid ambiguity
        return self.frame
        
    def get(self):
        if self.UseDefault():
            return self.defaultValue
        else:
            return TkIntEntry.get(self)
            
class TkDefaultFloatEntry(TkFloatEntry, TkDefaultMixin):

    def __init__(self, master, width, useDefault, defaultValue):
        TkDefaultMixin.__init__(self, master, useDefault, defaultValue)
        TkFloatEntry.__init__(self, self.frame, width)
        self.finish()        
        
    def tkWidget(self): # To avoid ambiguity
        return self.frame
        
    def get(self):
        if self.UseDefault():
            return self.defaultValue
        else:
            return TkFloatEntry.get(self)
            
            
            
class TkPopupSelector:
    def __init__(self, master, value2pop, pop2value, width):
    
        self.value2pop = value2pop
        self.pop2value = pop2value
        self.popupvalue = StringVar()
        self.popupvalue.set(self.pop2value.keys()[0]) # XXX first value as default 
        
        # XXX Uuughhh
        keys = self.value2pop.keys()
        keys.sort()
        pops = map(lambda x: value2pop[x], keys)
        #log.debug("pops = %s" % pops)
        args = (master, self.popupvalue) + tuple(pops)
        
        self.tkwidget = apply(OptionMenu, args)
        self.tkwidget.config(height=1, width=width)
        
    def tkWidget(self):
        return self.tkwidget
        
    def get(self):
        return self.pop2value[self.popupvalue.get()]
        
    def set(self, value):
        try:
            self.popupvalue.set(self.value2pop[value])
        except:
            self.popupvalue.set(self.pop2value.keys()[0]) # XXX first value as default       
            
    def select(self):    
        # Cant choose invalid value with popup
        pass
        
class TkStringPopupSelector:
    def __init__(self, master, strings):
    
        self.strings = strings
        self.popupvalue = StringVar()
        if len(self.strings) > 0:
            self.popupvalue.set(self.strings[0]) # XXX first value as default 
            
        width = max(map(len, self.strings))
        args = (master, self.popupvalue) + tuple(self.strings)
        self.tkwidget = apply(OptionMenu, args)
        self.tkwidget.config(height=1, width=width)
        
    def tkWidget(self):
        return self.tkwidget
        
    def get(self):
        return self.popupvalue.get()
        
    def set(self, value):
        try:
            self.popupvalue.set(value)
        except:
            self.popupvalue.set(self.strings[0]) # XXX first value as default       
            
    def select(self):    
        # Cant choose invalid value with popup
        pass
        
        
class TkColorSelector:
    def __init__(self, master, color='black'):
        #self.tkwidget = Button(master, width=8, command=self.editColor)
        self.tkwidget = Frame(master, height=18, width=60, relief=RIDGE, borderwidth=1)
        self.tkwidget.bind("<ButtonRelease-1>", self.editColor)
        self.set(color)
        
    def editColor(self, event):
        color = askcolor(self.color)[1]
        if color is not None:
            self.set(color)
            
    def tkWidget(self):
        return self.tkwidget
        
    def get(self):
        return self.color
        
    def set(self, value):
        self.color = value
        self.tkwidget.config(bg=self.color)
        
    def select(self):    
        # Cant choose invalid value with popup
        pass
        
        
        
class EditObjectAttributesDialog(tkSimpleDialog.Dialog):
    """ Creates an editable (pseudo-)inspector for a selected set of
        attributes of a given object
    
         - master : tk master widget
         - object : the object, whose attributes we want to edit
         - attr_names : a list of attr_names or a dictionary of attr_names and displayed Text
    
        By making use of Python 2.2's capability of subclassing built-in
        types such as ints, information about editing etc. is conveyed.
        An attr must have:
         - validate(value) method [return 1, if value is a valid new value for attr]
    
        The class of an attr can have the following mix-ins:
         - Popubable 
         - WithDefault 
    """
    
    def __init__(self, master, object, attr_names):
        self.object = object
        self.attr_names = attr_names
        self.edit = {}
        tkSimpleDialog.Dialog.__init__(self, master, "Edit: %s" % self.object.desc)
        
        
    def editWidget(self, master, object, attr_name):
        """ Create a widget capable of editing attr and insert attr's current value"""
        
        attr = object.__dict__[attr_name]
        attr_type = type(attr)
        widget = None
        default = isinstance(attr, WithDefault) # has a WithDefault mixin
        
        if isinstance(attr, Popupable):            
            widget = TkPopupSelector(master, attr.val2pop, attr.pop2val, attr.width)
            
        elif isinstance(attr, ValidatingBool):
            widget = TkBoolEntry(master)

        elif isinstance(attr, str):
        
            if default:
                widget = TkDefaultStringEntry(master, max(32, len(attr)), attr.useDefault, attr)
            else:
                widget = TkStringEntry(master, max(32, len(attr)))
                
        elif isinstance(attr, int):
        
            if default:
                widget = TkDefaultIntEntry(master, 6, attr.useDefault, attr)
            else:
                widget = TkIntEntry(master, 6)
                
        elif isinstance(attr, float):
        
            if default:
                widget = TkDefaultFloatEntry(master, 8, attr.useDefault, attr)
            else:
                widget = TkFloatEntry(master, 8)
                
        widget.set(attr)
        return widget
        
        
    def body(self, master):
        self.resizable(0,0)	
        
        # Header Zeile
        label = Label(master, text="Name", anchor=E)
        label.grid(row=0, column=0, padx=4, pady=3, sticky=E)
        label = Label(master, text="Value", anchor=W)
        label.grid(row=0, column=1, padx=4, pady=3, sticky=W)
        
        cur_row = 1

        if isinstance(self.attr_names, list):
            for attr in self.attr_names:
                label = Label(master, text="%s" % attr, anchor=E)
                label.grid(row=cur_row, column=0, padx=4, pady=3, sticky=E)

                self.edit[attr] = self.editWidget(master, self.object, attr)
                if self.edit[attr] != None:
                    self.edit[attr].tkWidget().grid(row=cur_row, column=1, padx=2, pady=1, sticky=W)

                cur_row = cur_row + 1
        else:
            for attr in self.attr_names.keys():
                label = Label(master, text="%s" % self.attr_names[attr], anchor=E)
                label.grid(row=cur_row, column=0, padx=4, pady=3, sticky=E)

                self.edit[attr] = self.editWidget(master, self.object, attr)
                if self.edit[attr] != None:
                    self.edit[attr].tkWidget().grid(row=cur_row, column=1, padx=2, pady=1, sticky=W)

                cur_row = cur_row + 1

    def validate(self):
        for attr_name in self.edit.keys():
            try:
            
                # In python 2.2 we can subclass attributes and add a validate method
                # to attributes
            
                value = self.edit[attr_name].get()
                
                if self.object.__dict__[attr_name].validate(value) == 0:
                    raise ValueError
                    
            except ValueError:
                msg = "Please enter a valid value for %s" % attr_name
                tkMessageBox.showwarning("Invalid Value", msg, parent=self)
                self.edit[attr_name].select()
                return 0
                
                # Everything is valid => set values
        for attr_name in self.edit.keys():            
            self.object.__dict__[attr_name] = typed_assign(self.object.__dict__[attr_name], self.edit[attr_name].get())
            
            if isinstance(self.object.__dict__[attr_name], WithDefault):
                self.object.__dict__[attr_name].useDefault = self.edit[attr_name].useDefault.get()
                
                
        return 1
        
        
        #-------------------------------------------------------------------------------
class WithDefault:
    """Mix-in for variables which have a default value"""
    
    def setDefault(self, useDefault, defaultValue):
        self.useDefault = useDefault
        self.defaultValue = defaultValue
        
    def validate(self, value):
    ##        if self.useDefault:
    ##            return 1
    ##        else:
    ##            return 1 # XXX How can I call a method of the class I am mixed too
        return 1
        
        
class Popupable:
    """Mix-in for variables which can be edited via a pop-up menu
       - val2pop : dict mapping value to string for pop up menu
       - pop2val: dict mapping pop up menu string to value
       - width: maximal string length in pop up
    """
    def setPopup(self, val2pop, pop2val = None, width = None):
        self.val2pop = val2pop
        self.pop2val = None
        self.width = None
        
        if pop2val == None:
            self.pop2val = {} # Private copy
            self.width = 0
            
            for val in val2pop.keys():
                pop = val2pop[val]
                self.width = max(len(pop), self.width)
                self.pop2val[pop] = val
        else:
            self.pop2val = pop2val
            self.width = width
            
    def validate(self, value):
        return 1
        
        ##class PopupableStr(str):
        ##    """Class for variables which can be edited via a pop-up menu
        ##       - values: array of values 
        ##       - width: maximal string length in pop up
        ##    """
        ##    def setPopup(self, values, width = None):
        
        ##        self.values = values
        ##        self.width = width
        
        ##        if width == None:
        ##            self.width = 0
        
        ##            for s in values:
        ##                self.width = max(len(s), self.width)
        
        ##    def validate(self, value):
        ##        return 1
        
        
class AlwaysValidate:
    """Mix-in for variables which always are valid"""
    def validate(self, value):
        return 1
        
        #-------------------------------------------------------------------------------
class ValidatingBool(int):
    """Editable replacement for bool"""
    def validate(self, value):
        if value is True or value is False:
            return 1

class ValidatingInt(int, AlwaysValidate):
    """Editable replacement for ints"""
    pass
    
class  ValidatingFloat(float, AlwaysValidate):
    """Editable replacement for floats"""
    pass
    
class  ValidatingString(str, AlwaysValidate):
    """Editable replacement for strings"""
    pass
    
class PopupableInt(int, Popupable):
    """A replacement for ints editable via a pop-up"""
    pass
    
class Probability(float):
    """An editable float taking values from [0,1]"""
    def validate(self, value):
        if 0.0 <= value and value <= 1.0:
            return 1
        else:
            return 0
            
class DefaultedInt(int, WithDefault):
    """An editable int with a default value"""    
    pass
    
class DefaultedFloat(float, WithDefault):
    """An editable float with a default value"""    
    pass
    
class DefaultedString(str, WithDefault):
    """An editable strinf with a default value"""    
    pass
    
    
    #======================================================================
    #
    # Demo:
    #
class TkTestFrame(Frame):

    def __init__(self, parent=None):
        Frame.__init__(self,parent)
        Pack.config(self)
        self.createWidgets()
        
        self.desc = ValidatingString("The TkTestFrame")
        self.x = DefaultedInt(1)
        self.x.setDefault(1, 122)
        self.y = ValidatingFloat(2.33)
        self.choose = PopupableInt(3)
        self.pop2val = {"aaa":1, "xxx":2, "sss":3}
        self.val2pop = {1:"aaa", 2:"xxx", 3:"sss"}
        self.choose.setPopup(self.val2pop, self.pop2val, 5)
        
    def createWidgets(self):
        self.QUIT = Button(self, text='QUIT', foreground='red', 
                           command=self.quit)
        self.QUIT.pack(side=LEFT)
        self.About = Button(self, text='Preferences', foreground='red', 
                           command=self.About)
        self.About.pack(side=LEFT)
        
        
    def About(self):
        aboutBox = EditObjectAttributesDialog(self.master, self, ['desc', 'x', 'y', 'choose'])
        del self.pop2val["aaa"]
        del self.val2pop[1]
        aboutBox = EditObjectAttributesDialog(self.master, self, ['desc', 'x', 'y', 'choose'])
        
if __name__ == '__main__':
    app = TkTestFrame()
    app.mainloop()
    
    



################################################################################
#
#       This file is part of Gato (Graph Animation Toolbox) 
#
#	file:   GatoFile.py
#	author: Achim Gaedke (achim.gaedke@zpr.uni-koeln.de)
#
#       Copyright (C) 1998-2015, Alexander Schliep, Winfried Hochstaettler and 
#       Copyright 1998-2001 ZAIK/ZPR, Universitaet zu Koeln
#                                   
#       Contact: alexander@schliep.org, winfried.hochstaettler@fernuni-hagen.de             
#       Information: http://gato.sf.net
#
#       This library is free software; you can redistribute it and/or
#       modify it under the terms of the GNU Library General Public
#       License as published by the Free Software Foundation; either
#       version 2 of the License, or (at your option) any later version.
#
#       This library is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#       Library General Public License for more details.
#
#       You should have received a copy of the GNU Library General Public
#       License along with this library; if not, write to the Free
#       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#       This file is version $Revision: 670 $ 
#                       from $Date: 2015-01-13 16:04:11 -0500 (Tue, 13 Jan 2015) $
#             last change by $Author: schliep $.
#
################################################################################
import os
import os.path
import shutil
import sys
import string
import re
import Tkinter
import tkSimpleDialog
import tkFileDialog
import traceback
try:
    import _winreg
except ImportError:
    # we are not on a windows system
    pass
    
    
gatoMimeType="application/gato"
gatoFileExtension="gato"
gatoDescription="gato graph/algorithm tool"

class ConfigurationException(Exception):
    """
    to report configuration errors
    """
    def __init__(self,message):
        Exception.__init__(self)
        self.message=message
        
    def __repr__(self):
        if hasattr(self,"message"):
            return "%s: %s"%(self.__class__.__name__,self.message)
        else:
            return self.__class__.__name__
            
class configureOS:
    """
    system configuration for gato file type support
    """
    
    def __init__(self,DialogMaster=None):
        """
        creates configurator object
        """
        # find script location
        self.myExecutable=os.path.abspath(sys.argv[0])
        self.DialogMaster=DialogMaster
        
    def check(self):
        """
        check, if already configured, returns true, if this executable is installed
        """
        raise ConfigurationException("base class method should not be called")
        
    def askUserInstall(self):
        """
        ask user, if she/he likes configuration
        """
        raise ConfigurationException("base class method should not be called")
        
    def askUserUninstall(self):
        """
        ask user, if she/he likes uninstallation
        """
        raise ConfigurationException("base class method should not be called")
        
    def configureSystem(self):
        """
        do system configuration
        """
        raise ConfigurationException("base class method should not be called")
        
    def unconfigureSystem(self):
        """
        do system unconfiguration
        """
        raise ConfigurationException("base class method should not be called")
        
    def runInstall(self):
        if not self.check() and self.askUserInstall():
            self.configureSystem()
            
    def runUninstall(self):
        if self.check() and self.askUserUninstall():
            self.unconfigureSystem()
            
    def installBinary(self,path):
        """
        installs this binary to another place
        """
        # test, if it is our place...
        if not os.path.exists(path):
            os.makedirs(path)
        newLocation=os.path.join(path,os.path.basename(self.myExecutable))
        if os.path.exists(newLocation):
            os.remove(newLocation)
        shutil.copy2(self.myExecutable,newLocation)
        self.myExecutable=os.path.abspath(newLocation)
        
class configureUnsupported(configureOS):
    """
    some tips and exit
    """
    check=configureSystem=askUserInstall=lambda x:None
    
    def runInstall(self):
        """
        """
        print "unsupported operating system %s"%sys.platform
        
class configureUNIX(configureOS):
    """
    linux configuration
    expands mailcap file
    """
    
    def __init__(self,DialogMaster=None):
        """
        find script location, configuration file location...
        """
        configureOS.__init__(self,DialogMaster)
        
        # find mailcap file
        if not os.environ.has_key("HOME"):
            raise ConfigurationException("could not determine home directory")
        self.mailcapFile=os.path.join(os.environ["HOME"],".mailcap")
        self.mime_typesFile=os.path.join(os.environ["HOME"],".mime.types")
        
    def check(self):
        """
        """
        self.cache_check_mailcap=self.check_mailcap()
        self.cache_check_mime_types=self.check_mime_types()
        return self.cache_check_mailcap and self.cache_check_mime_types
        
    def check_mailcap(self):
        """        
        """
        if not os.access(self.mailcapFile,os.R_OK):
            return 0
            
        mailcap=file(self.mailcapFile,"r")
        while 1:
            line=mailcap.readline()
            # line end
            if not line:
                break
                
                # skip comments
            if line[0]=="#":
                continue
               
            line=line.strip()
            # skip whitespace
            if len(line) == 0:
                continue                                      
            while line[-1]=="\\":
                line=line[:-1]+mailcap.readline().rstrip()
            entries=line.split(";")
            (mimeType,viewCommand)=map(string.strip,entries[:2])
            
            if (mimeType==gatoMimeType and
                os.path.exists(viewCommand.split(" ")[0]) and
                os.path.samefile(viewCommand.split(" ")[0],self.myExecutable)):
                return 1
                
        return 0
        
    def check_mime_types(self):
        """
        """
        if not os.access(self.mime_typesFile,os.R_OK):
            return 0
            
        mime_types=file(self.mime_typesFile,"r")
        while 1:
            line=mime_types.readline()
            # line end
            if not line:
                break
                
                # skip comments
            if line[0]=="#":
                continue
                
                #catenate lines
            line=line.strip()
            while line[-1]=="\\":
                newLine=mime_types.readline()
                if newLine=="":
                    break
                line=line[:-1]+newLine.rstrip()
                
            mime_dict={}
            # format: key=value or key="value"
            kv_pair=re.compile("(\w+)=((?:\"[^\"]*\")|(?:[^\"\s]*))")
            kv_pairs=kv_pair.findall(line)
            for kv in kv_pairs:
                (key,value)=kv
                if value[0]=="\"":
                    value=value[1:-1]
                mime_dict[key]=value
            if mime_dict.get("type")=="application/gato" and mime_dict.get("exts")=="gato":
                return 1
                
        return 0
        
    class askUserInstallDialog(tkSimpleDialog.Dialog):
        """
        dialog for system file manipulation of linux systems
        """
        
        def __init__(self,master,title=None):
            if not title:
                title="System Configuration"
            tkSimpleDialog.Dialog.__init__(self,master,title)
            
        def body(self, master):
            """
            """
            row=0
            # install prefix question
            Tkinter.Label(master, text="Install Gato to another place?").grid(row=row,column=0)
            self.installQ = Tkinter.IntVar()
            self.installQ.set(1)
            Tkinter.Radiobutton(master, text="Yes", variable=self.installQ, value=1,
                                command=self.enablePathEntry).grid(row=row,column=1)
            Tkinter.Radiobutton(master, text="No",  variable=self.installQ, value=2,
                                command=self.disablePathEntry).grid(row=row,column=2)
            
            # install location
            row+=1
            self.installP=Tkinter.Entry(master)
            self.installP.insert(0,os.path.expanduser("~/bin"))
            self.installP.rowLocation=row
            self.installP.colLocation=0
            self.installP.grid(row=row,column=0,sticky=Tkinter.EW)
            self.searchP=Tkinter.Button(master,text="search...",
                                        command=self.askInstallPrefix,
                                        pady=0)
            self.searchP.grid(row=row,column=1,columnspan=2)
            
            # mime type reg?
            row+=1
            Tkinter.Label(master, text="add gato mime type").grid(row=row,column=0)
            self.mimeQ = Tkinter.IntVar()
            self.mimeQ.set(1)
            Tkinter.Radiobutton(master, text="Yes", variable=self.mimeQ,
                                value=1).grid(row=row,column=1)
            Tkinter.Radiobutton(master, text="No",  variable=self.mimeQ,
                                value=2).grid(row=row,column=2)
            
            # mime type reg?
            row+=1
            Tkinter.Label(master, text="add gato to .mailcap").grid(row=row,column=0)
            self.mailcapQ = Tkinter.IntVar()
            self.mailcapQ.set(1)
            Tkinter.Radiobutton(master, text="Yes", variable=self.mailcapQ,
                                value=1).grid(row=row,column=1)
            Tkinter.Radiobutton(master, text="No",  variable=self.mailcapQ,
                                value=2).grid(row=row,column=2)
            
        def enablePathEntry(self):
            self.installP.grid(row=self.installP.rowLocation,column=self.installP.colLocation,sticky=Tkinter.EW)
            self.searchP["state"]=Tkinter.NORMAL
            
        def disablePathEntry(self):
            self.installP.grid_forget()
            self.searchP["state"]=Tkinter.DISABLED
            
        def askInstallPrefix(self):
            """
            command for search button
            """
            initDir=self.installP.get()
            newPrefix=""
            if os.path.exists(initDir):
                newPrefix=tkFileDialog.askdirectory(parent=self,initialdir=initDir,mustexist=1)
            else:
                newPrefix=tkFileDialog.askdirectory(parent=self,mustexist=1)                
            if newPrefix:
                self.installP.delete(0, Tkinter.END)
                self.installP.insert(0, newPrefix)
                
        def apply(self):
            """
            return result as dictionary
            """
            self.result={}
            if (hasattr(self,"installP") and 
                hasattr(self,"installQ") and
                (self.installQ.get()==1)):
                self.result["installTo"]=self.installP.get()
            else:
                self.result["installTo"]=None
            if hasattr(self,"mimeQ"):
                self.result["mimeQ"]=self.mimeQ.get()==1
            if hasattr(self,"mailcapQ"):
                self.result["mailcapQ"]=self.mailcapQ.get()==1
                
    def askUserInstall(self):
        """
        """
        self.askedUser=self.askUserInstallDialog(self.DialogMaster).result
        return self.askedUser
        
    class askUserUninstallDialog(tkSimpleDialog.Dialog):
        """
        dialog for system file manipulation of linux systems
        """
        def __init__(self,master,title=None):
            if not title:
                title="System Configuration"
            tkSimpleDialog.Dialog.__init__(self,master,title)
            
        def body(self, master):
            """
            Are you sure...? Dialog...
            """
            row=0
            # install prefix question
            Tkinter.Label(master,
                          text="Install Gato from\n%s ?"%self.myExecutable
                          ).grid(row=row,column=0)
            
        def apply(self):
            """
            return result as dictionary
            """
            self.result={}
            
    def askUserUninstall(self):
        self.askedUser=self.askUserUnnstallDialog(self.DialogMaster).result
        return self.askedUser
        
    def configureSystem(self):
        """
        """
        if self.askedUser.get("installTo",0):
            self.installBinary(self.askedUser["installTo"])
        if not self.cache_check_mailcap and self.askedUser.get("mailcapQ",0):
            self.configure_mailcap()
        if not self.cache_check_mime_types and self.askedUser.get("mimeQ",0):
            self.configure_mime_types()
            
    def configure_mailcap(self):
        """
        append my mime type to .mailcap
        """
        savedLines=""
        # if this file exists
        if os.access(self.mailcapFile,os.R_OK):
            mailcap=file(self.mailcapFile,"r")
            
            while 1:
                line=mailcap.readline()
                savedLine=line[:]
                # line end
                if not line:
                    break
                    
                    # skip comments
                if line[0]=="#":
                    savedLines+=line
                    continue
                    
                line=line.strip()
                continuedLines=[]
                while line[-1]=="\\":
                    continuedLines.append(savedLine)
                    savedLine=mailcap.readline()
                    line=line[:-1]+savedLine.rstrip()
                entries=line.split(";")
                (mimeType,viewCommand)=map(string.strip,entries[:2])
                
                # skip my old entry
                if mimeType==gatoMimeType:
                    continue
                    
                savedLines+=string.join(continuedLines,'')+savedLine
            mailcap.close()
            # open for write access
        mailcap=file(self.mailcapFile,"w")
        mailcap.write(savedLines)
        mailcap.write("%s;%s \"%%s\"\n"%(gatoMimeType,self.myExecutable))
        mailcap.close()
        
    def configure_mime_types(self):
        """
        append my information to .mime.types
        """
        savedLines=""
        # if this file exists
        if os.access(self.mime_typesFile,os.R_OK):
            mime_types=file(self.mime_typesFile,"r")
            
            while 1:
                line=mime_types.readline()
                savedLine=line[:]
                # line end
                if not line:
                    break
                    
                    # skip comments
                if line[0]=="#":
                    savedLines+=line
                    continue
                    
                line=line.strip()
                continuedLines=[]
                while line[-1]=="\\":
                    continuedLines.append(savedLine)
                    savedLine=mime_types.readline()
                    line=line[:-1]+savedLine.rstrip()
                entries=line.split(";")
                
                mime_dict={}
                # format: key=value or key="value"
                kv_pair=re.compile("(\w+)=((?:\"[^\"]*\")|(?:[^\"\s]*))")
                kv_pairs=kv_pair.findall(line)
                for kv in kv_pairs:
                    (key,value)=kv
                    if value[0]=="\"":
                        value=value[1:-1]
                    mime_dict[key]=value
                if (mime_dict.get("type")==gatoMimeType and
                    mime_dict.get("exts")==gatoFileExtension):
                    continue
                    
                savedLines+=string.join(continuedLines,'')+savedLine
            mime_types.close()
        else:
            # fake Netscape MIME file
            savedLines+="#--Netscape Communications Corporation MIME Information\n"
            # open for write access
        mime_types=file(self.mime_typesFile,"w")
        mime_types.write(savedLines)
        mime_types.write("# inserted by gato SystemConfiguration Module\n")
        mime_types.write("type=%s  \\\ndesc=\"%s\"  \\\nexts=\"%s\"\n"%
                         (gatoMimeType,gatoDescription,gatoFileExtension))
        mime_types.close()
        
        
class configureLinux(configureUNIX):
    """
    """
    def __init__(self, DialogMaster=None):
        if sys.platform[:5]!='linux':
            raise ConfigurationException("tried to instantiate %s on %s"%
                                         (self.__class__.__name__,sys.platform))
        configureUNIX.__init__(self,DialogMaster)
        
class configureSUNOS(configureUNIX):
    """
    """
    def __init__(self, DialogMaster=None):
        if sys.platform[:5]!='sunos':
            raise ConfigurationException("tried to instantiate %s on %s"%
                                         (self.__class__.__name__,sys.platform))
        configureUNIX.__init__(self,DialogMaster)
        
class configureWindows(configureOS):
    """
    Configuration module for windows
    contaminates the windows registry with our
    extension, program and mime type
    """
    
    def __init__(self,DialogMaster=None):
        """
        find script location...
        """
        configureOS.__init__(self,DialogMaster)
        
    def check(self):
        """
        """
        self.ClassesSection=self.findWritableClassesSection()
        return 0
        
    class askUserInstallDialog(tkSimpleDialog.Dialog):
        """
        """
        def __init__(self,master,title=None):
            if not title:
                title="System Configuration"
            tkSimpleDialog.Dialog.__init__(self,master,title)
            
        def body(self, master):
            """
            dialog body
            """
            row=0
            # install question
            Tkinter.Label(master, text="Install Gato.exe to another place?").grid(row=row,column=0)
            self.installQ = Tkinter.IntVar()
            self.installQ.set(1)
            Tkinter.Radiobutton(master, text="Yes", variable=self.installQ, value=1, command=self.enablePathEntry).grid(row=row,column=1)
            Tkinter.Radiobutton(master, text="No",  variable=self.installQ, value=2, command=self.disablePathEntry).grid(row=row,column=2)
            # install location
            row+=1
            self.installP=Tkinter.Entry(master)
            self.installP.insert(0,"C:\\Gato\\")
            self.installP.rowLocation=row
            self.installP.colLocation=0
            self.installP.grid(row=row,column=0,sticky=Tkinter.EW)
            self.searchP=Tkinter.Button(master,text="search...",command=self.askInstallPrefix, pady=0)
            self.searchP.grid(row=row,column=1,columnspan=2)
            # extension question
            row+=1
            self.extensionQ = Tkinter.IntVar()
            self.extensionQ.set(1)
            Tkinter.Label(master, text="Create bindings to .%s file extensions:"%gatoFileExtension).grid(row=row)
            Tkinter.Radiobutton(master, text="Yes", variable=self.extensionQ, value=1).grid(row=row,column=1)
            Tkinter.Radiobutton(master, text="No",  variable=self.extensionQ, value=2).grid(row=row,column=2)
            # MIME type question
            row+=1
            self.mimeQ = Tkinter.IntVar()
            self.mimeQ.set(1)
            Tkinter.Label(master, text="Register MIME type %s:"%gatoMimeType).grid(row=row)
            Tkinter.Radiobutton(master, text="Yes", variable=self.mimeQ, value=1).grid(row=row,column=1)
            Tkinter.Radiobutton(master, text="No",  variable=self.mimeQ, value=2).grid(row=row,column=2)
            
        def apply(self):
            """
            return result as dictionary
            """
            self.result={}
            if (hasattr(self,"installP") and 
                hasattr(self,"installQ") and
                (self.installQ.get()==1)):
                self.result["installTo"]=self.installP.get()
            else:
                self.result["installTo"]=None
            if hasattr(self,"mimeQ"):
                self.result["mimeQ"]=self.mimeQ.get()==1
            if hasattr(self,"extensionsQ"):
                self.result["extensionsQ"]=self.extensionsQ.get()==1
                
        def enablePathEntry(self):
            self.installP.grid(row=self.installP.rowLocation,column=self.installP.colLocation,sticky=Tkinter.EW)
            self.searchP["state"]=Tkinter.NORMAL
            
        def disablePathEntry(self):
            self.installP.grid_forget()
            self.searchP["state"]=Tkinter.DISABLED
            
        def askInstallPrefix(self):
            """
            command for search button
            """
            initDir=self.installP.get()
            newPrefix=""
            if os.path.exists(initDir):
                newPrefix=tkFileDialog.askdirectory(parent=self,initialdir=initDir,mustexist=1)
            else:
                newPrefix=tkFileDialog.askdirectory(parent=self,mustexist=1)                
            if newPrefix:
                self.installP.delete(0, Tkinter.END)
                self.installP.insert(0, newPrefix)
                
    def askUserInstall(self):
        """
        """
        self.askedUser=self.askUserInstallDialog(self.DialogMaster).result
        return self.askedUser
        
    def configureSystem(self):
        """
        do the system configuration
        """
        if self.askedUser.get("installTo",0):
            self.installBinary(self.askedUser["installTo"])
        if self.askedUser.get("extensionQ",0) or self.askedUser.get("mimeQ",0):
            self.insertGatoEntries(self.ClassesSection)
            
    def printGatoEntries(self):
        """
        start reading...
        """
        reader=_winreg.ConnectRegistry(None,_winreg.HKEY_CLASSES_ROOT)
        # get Gato.File section
        GatoFileHandle=None
        try:
            GatoFileHandle=_winreg.OpenKey(reader,"Gato.File")
        except WindowsError:
            print "Could not find Gato.File section"
        else:
            print "found Gato.File section:"
            self.printSubRegistry(GatoFileHandle)
            
            # get Gato FileExtension Section
        GatoExtensionHandle=None
        try:
            GatoExtensionHandle=_winreg.OpenKey(reader,"."+gatoFileExtension)
        except WindowsError:
            print "could not find the file extension .%s"%gatoFileExtension
        else:
            print "found gato's extension"
            self.printSubRegistry(GatoExtensionHandle)
            
            # get Gato mime Type section
        GatoMimeHandleGatoExt=None
        try:
            GatoMimeHandle=_winreg.OpenKey(reader,"MIME")
            GatoMimeHandleDatabase=_winreg.OpenKey(GatoMimeHandle,
                                                    "Database")
            GatoMimeHandleContentType=_winreg.OpenKey(GatoMimeHandleDatabase,
                                                      "Content Type")
            GatoMimeHandleGatoExt=_winreg.OpenKey(GatoMimeHandleContentType,
                                                  gatoMimeType)
        except WindowsError:
            print "could not find mime type: %s"%gatoMimeType
        else:
            print "found %s mime type"%gatoMimeType
            self.printSubRegistry(GatoMimeHandleGatoExt)
            
    def printSubRegistry(self,key,indent=""):
        """
        print all information of a subkey
        """
        subkeyNo,valueNo,lastMod=_winreg.QueryInfoKey(key)
        for i in range(valueNo):
            print indent,_winreg.EnumValue(key, i)
        for i in range(subkeyNo):
            subkeyName=_winreg.EnumKey(key,i)
            subkey=_winreg.OpenKey(key,subkeyName)
            print indent,subkeyName
            self.printSubRegistry(subkey,indent+"  ")
            
    def findWritableClassesSection(self):
        # first try in HKEY_CLASSES_ROOT
        try:
            writer=_winreg.ConnectRegistry(None,_winreg.HKEY_CLASSES_ROOT)
            GatoFileTestHandle=_winreg.CreateKey(writer,"Gato.File.Test")
            _winreg.DeleteKey(writer,"Gato.File.Test")
            return writer
        except WindowsError:
            # print "could not access HKEY_CLASSES_ROOT/Gato.File"
            # self.traceback.print_exc()
            pass
            # next try...
        try:
            writer=_winreg.ConnectRegistry(None,_winreg.HKEY_CURRENT_USER)
            SoftwareSection=_winreg.OpenKey(writer,"Software")
            ClassesSection=_winreg.OpenKey(SoftwareSection,"Classes",0,_winreg.KEY_SET_VALUE)
            GatoFileTestHandle=_winreg.CreateKey(ClassesSection,"Gato.File.Test")
            _winreg.DeleteKey(ClassesSection,"Gato.File.Test")
            return ClassesSection
        except WindowsError:
            # print "could not access HKEY_CURRENT_USER/Software/Classes Section"
            # self.traceback.print_exc()
            return None
            
    def insertGatoEntries(self,ClassesSection):
        """
        updates registry database
        """
        # update Gato.File section
        try:
            GatoFileHandle=_winreg.CreateKey(ClassesSection,"Gato.File")
            _winreg.SetValueEx(GatoFileHandle,"",0,_winreg.REG_SZ,"Gato.File")
        except WindowsError:
            print "Could not create/update the Gato.File section"
            self.traceback.print_exc()
            
            # update Gato.File's subsections
        try:
            GatoShellHandle=_winreg.CreateKey(GatoFileHandle,"shell")
            GatoOpenHandle=_winreg.CreateKey(GatoShellHandle,"open")
            GatoOpenCommandHandle=_winreg.CreateKey(GatoOpenHandle,"command")
            _winreg.SetValueEx(GatoOpenCommandHandle,"",0,_winreg.REG_SZ,self.myExecutable+' "%1"')
        except WindowsError:
            print "could not install open command for gato"
            self.traceback.print_exc()
            
            # update .gato section    
        try:
            GatoExtensionHandle=_winreg.CreateKey(ClassesSection,"."+gatoFileExtension)
            _winreg.SetValueEx(GatoExtensionHandle,"",0,_winreg.REG_SZ,"Gato.File")
            _winreg.SetValueEx(GatoExtensionHandle,"Content Type",0,_winreg.REG_SZ,gatoMimeType)
        except WindowsError:
            print "could not create/update FileExtension section"
            self.traceback.print_exc()
            
            # access MIME Database section
        MimeContentTypeSection=None
        try:
            MimeSection=_winreg.CreateKey(ClassesSection,"MIME")
            MimeDatabaseSection=_winreg.CreateKey(MimeSection,"Database")
            MimeContentTypeSection=_winreg.CreateKey(MimeDatabaseSection,"Content Type")
        except WindowsError:
            print "could not access MIME Content Type database"
            self.traceback.print_exc()
            return
            # update gato's mime type
        try:
            GatoMimeContentTypeHandle=_winreg.CreateKey(MimeContentTypeSection,gatoMimeType)
            _winreg.SetValueEx(GatoMimeContentTypeHandle,"Extension",0,
                                    _winreg.REG_SZ,"."+gatoFileExtension)
        except WindowsError:
            print "could not update the gato MIME section"
            self.traceback.print_exc()
            
class GatoInstaller:
    """
    A instance of this class reflects the installation capabilities and status.
    It should be instantiated only once.
    """
    instanceCounter=0
    
    def __init__(self,disabled=0,menu=None,index=Tkinter.END):
        """
        gets system configurator and checks state
        """
        # assure singleton
        if GatoInstaller.instanceCounter!=0:
            raise ConfigurationException("class GatoInstaller should be instantiated once.")
        GatoInstaller.instanceCounter+=1
        self.disabled=disabled
        self.menu=menu
        self.index=index
        # set if no menu item should appear
        if self.disabled:
            return
        self.SysConfig=self.getConfigurator()
        self.state=self.SysConfig.check()
        
    def __del__(self):
        """
        clean up and decrement
        """
        GatoInstaller.instanceCounter-=1
        
    def enable(self):
        self.disabled=0
        self.SysConfig=getConfigurator()
        self.state=self.SysConfig.check()
        if self.menu is not None:
            self.insertMenuEntry()
            
    def disable(self):
        self.disabled=1
        self.removeMenuEntry()
        
    def addMenuEntry(self,menu,index=Tkinter.END):
        self.menu=menu
        self.index=menu.index(index) # make absolute position
        if self.disabled:
            return
        self.insertMenuEntry()
        
    def insertMenuEntry(self):
        if self.menu is None:
            return
        self.menu.insert_command(self.index)
        self.configureMenuEntry()
        
    def removeMenuEntry(self):
        if self.menu is None:
            return
        self.menu.delete(self.index)
        
    def configureMenuEntry(self):
        if self.menu is None:
            return
        if self.state:
            self.menu.entryconfig(self.index,
                                  command=self.uninstallCommand,
                                  label='Uninstall Gato'
                                  )
        else:
            self.menu.entryconfig(self.index,
                                  command=self.installCommand,
                                  label='Install Gato...')
            
    def installCommand(self):
        self.SysConfig.runInstall()
        self.state=self.SysConfig.check()
        self.configureMenuEntry()
        
    def uninstallCommand(self):
        print "uninstall"
        self.state=self.SysConfig.check()
        self.configureMenuEntry()
        
    def getConfigurator(self):
        if sys.platform[:5]=="linux":
            return configureLinux()
        elif sys.platform=="win32":
            return configureWindows()
        elif sys.platform[:5]=="sunos":
            return configureSUNOS()
        else:
            return configureUnsupported()
            
if __name__=="__main__":
    i=GatoInstaller()
    
    
    
    
